{------------------------------------------------------------------------------}
{ TWAIN Sample Source                                                          }
{ Copyright (C) '91-'93 TWAIN Working Group:                                   }
{ Aldus, Caere, Eastman-Kodak, Logitech,                                       }
{ Hewlett-Packard Corporations.                                                }
{ All rights reserved.                                                         }
{                                                                              }
{ This source code was converted from C to Turbo Pascal for Windoes by         }
{                                                                              }
{   MCM Design                                                                 }
{   Marc Christopher Martin                                                    }
{   Australiensvej 34                                                          }
{   2100 Copenhagen OE                                                         }
{   Denmark                                                                    }
{                                                                              }
{   e-mail support@mcm-design.dk                                               }
{                                                                              }
{------------------------------------------------------------------------------}

unit twd_prot;

{------------------------------------------------------------------------------}
{$A+} { Align data Byte = $A-, Word = $A+.                                     }
{$F+} { Force FAR call.                                                        }
{$N-} { CoProcessor code generation.                                           }
{$R-} { Range check.                                                           }
{$W-} { Protected mode only $W-, Real mode stack frame support $W+.            }
{$X+} { Extended syntax (When on, function can be used as procedures).         }
{------------------------------------------------------------------------------}

interface

uses WinTypes,
     Dialogs,
     twain,    { req. for TWAIN types.                                         }
     twd_type, { function prototypes for Source.                               }
     twd_com;  { common lib of container routines.                             }


{ To allow compatibility between 32 and 16 bit _llseek.                        }
{$IFNDEF WIN32}
   const FILE_BEGIN   = 0;
         FILE_CURRENT = 1;
         FILE_END     = 2;
{$ENDIF}


{ Local function prototypes -- need to move these to a header.                 }

{$IFNDEF WIN32}
{ 16 Bit code.                                                                 }
function  CopyFileEx(Source : PChar; Dest : Pchar; Notify : bool) : bool;
{$ENDIF}

function  SetSourceControl : bool;

procedure LogMessage       (Msg     : Pchar);

function Dispatch          (PTWMsg  : PTWMessage;
                            pMapper : PMsgMap;
                            DAT     : TW_UINT16)         : TW_UINT16;

function IdentityMsg       (PTWMsg  : PTWMessage)        : TW_UINT16;

function InterfaceMsg      (PTWMsg  : PTWMessage)        : TW_UINT16;

function CapabilityMsg     (PTWMsg  : PTWMessage)        : TW_UINT16;

function StatusMsg         (PTWMsg  : PTWMessage)        : TW_UINT16;

function PendingXferMsg    (PTWMsg  : PTWMessage)        : TW_UINT16;

function SetupMemXferMsg   (PTWMsg  : PTWMessage)        : TW_UINT16;

function SetupFileXferMsg  (PTWMsg  : PTWMessage)        : TW_UINT16;

function XferGroupMsg      (PTWMsg  : PTWMessage)        : TW_UINT16;

function Palette8Msg       (PTWMsg  : PTWMessage)        : TW_UINT16;

function GrayResponseMsg   (PTWMsg  : PTWMessage)        : TW_UINT16;

function RGBResponseMsg    (PTWMsg  : PTWMessage)        : TW_UINT16;

function CIEColorMsg       (PTWMsg  : PTWMessage)        : TW_UINT16;

function JPEGCompressionMsg(PTWMsg  : PTWMessage)        : TW_UINT16;

function ImageInfoMsg      (PTWMsg  : PTWMessage)        : TW_UINT16;

function ImageLayoutMsg    (PTWMsg  : PTWMessage)        : TW_UINT16;

function ImageMemXferMsg   (PTWMsg  : PTWMessage)        : TW_UINT16;

function ImageFileXferMsg  (PTWMsg  : PTWMessage)        : TW_UINT16;

function ImageNativeXferMsg(PTWMsg  : PTWMessage)        : TW_UINT16;

function EnableDS          (pUI     : pTW_USERINTERFACE) : TW_UINT16;

function DisableDS         (pUI     : pTW_USERINTERFACE) : TW_UINT16;

function OpenDS            (PTWMsg  : PTWMessage)        : TW_UINT16;

function CloseDS                                         : TW_UINT16;

procedure NotifyXferReady;

procedure NotifyCloseDSReq;

procedure DS_Error         (state   : PChar;
                            code    : integer);

procedure ShowTW_ID        (hAppWnd : hWnd;
                            twID    : TW_IDENTITY;
                            Text    : PChar);

function IsPaletteGray : boolean;


procedure FlipBitmap(pBmp     : Pointer;
                     pBmpInfo : PBitmapInfo);

var SourceControl : array[0..MAX_TWPATH] of char;

{ Global of the app's window handle.                                           }
var hAppWnd       : hWnd;             { Global handle to applications window.  }
    DSStatus      : TW_STATUS;        { Status structure for errors.           }

    FileName      : array[0..512] of char;
    hPal          : TW_PALETTE8;


const DSState       : integer = STATE3; { Global state machine place holder.   }
      FeederEnabled : bool = FALSE;
      ControlMap    : array[0..8] of TMsgMapper
                    = ((ID    : DAT_IDENTITY;        pFunc : @IdentityMsg),
                       (ID    : DAT_USERINTERFACE;   pFunc : @InterfaceMsg),
                       (ID    : DAT_CAPABILITY;      pFunc : @CapabilityMsg),
                       (ID    : DAT_STATUS;          pFunc : @StatusMsg),
                       (ID    : DAT_PENDINGXFERS;    pFunc : @PendingXferMsg),
                       (ID    : DAT_SETUPMEMXFER;    pFunc : @SetupMemXferMsg),
                       (ID    : DAT_SETUPFILEXFER;   pFunc : @SetupFileXferMsg),
                       (ID    : DAT_XFERGROUP;       pFunc : @XferGroupMsg),
                       { THIS MUST BE LAST                                     }
                       (ID    : 0;                   pFunc : Nil));

      ImageMap      : array[0..10] of TMsgMapper
                    = ((ID    : DAT_IMAGEINFO;       pFunc : @ImageInfoMsg),
                       (ID    : DAT_IMAGELAYOUT;     pFunc : @ImageLayoutMsg),
                       (ID    : DAT_IMAGEMEMXFER;    pFunc : @ImageMemXferMsg),
                       (ID    : DAT_IMAGENATIVEXFER; pFunc : @ImageNativeXferMsg),
                       (ID    : DAT_IMAGEFILEXFER;   pFunc : @ImageFileXferMsg),
                       (ID    : DAT_PALETTE8;        pFunc : @Palette8Msg),
                       (ID    : DAT_GRAYRESPONSE;    pFunc : @GrayResponseMsg),
                       (ID    : DAT_RGBRESPONSE;     pFunc : @RGBResponseMsg),
                       (ID    : DAT_CIECOLOR;	     pFunc : @CIEColorMsg),
                       (ID    : DAT_JPEGCOMPRESSION; pFunc : @JPEGCompressionMsg),
                       { THIS MUST BE LAST.                                    }
                       (ID    : 0;                   pFunc : Nil));


const { Supported Capabilities.                                                }
      NUM_SUPPORTEDCAPS = 6;
      SupportedTypes : array[0..NUM_SUPPORTEDCAPS-1] of TW_UINT16
                     = (ICAP_XFERMECH,
                        ICAP_UNITS,
                        ICAP_PIXELTYPE,
                        CAP_XFERCOUNT,
                        ICAP_XRESOLUTION,
                        ICAP_YRESOLUTION);

      SupportedCaps : TW_ARRAY
                    = (ItemType : TWTY_UINT16;
                       NumItems : NUM_SUPPORTEDCAPS);


      { Application changed PixelType.                                         }
      AppChangedPix : boolean = False;

      { Transfer image using Buffers.                                          }
      MinBufSize    : integer = 0;
      MaxBufSize    : integer = 0;

      { PixelType (Required type is Emuneration).                              }
      {$IFDEF TCITWAIN}
      NUM_PIXTYPES  = 2;
      PixelType     : array[0..NUM_PIXTYPES-1] of TW_UINT16
                    = (TWPT_RGB, TWPT_GRAY);
      {$ELSE}
      NUM_PIXTYPES  = 4;
      PixelType     : array[0..NUM_PIXTYPES-1] of TW_UINT16
                    = (TWPT_RGB, TWPT_BW, TWPT_GRAY, TWPT_PALETTE);
      {$ENDIF}
      DEF_PIXTYPE   = 0;
      PixelEnum     : TW_ENUMERATION
                    = (ItemType     : TWTY_INT16;
                       NumItems     : NUM_PIXTYPES;
                       CurrentIndex : DEF_PIXTYPE;
                       DefaultIndex : DEF_PIXTYPE);


      { XferTypes (Required type is Enumeration).                              }
      NUM_XFERTYPES = 3;
      DEF_XFERTYPE  = 0;
      XferType      : array[0..NUM_XFERTYPES-1] of TW_UINT16
                    = (TWSX_NATIVE, TWSX_FILE, TWSX_MEMORY);

      XferEnum      : TW_ENUMERATION
                    = (ItemType     : TWTY_INT16;
                       NumItems     : NUM_XFERTYPES;
                       CurrentIndex : DEF_XFERTYPE;
                       DefaultIndex : DEF_XFERTYPE);


      { FileType transfer.                                                     }
      NUM_FILETYPES = 1;
      DEF_FILETYPE  = 0;
      FileType      : array[0..NUM_FILETYPES-1] of TW_UINT16
                    = (TWFF_BMP);
                    { This example only supports Bitmap transfer.              }
                    { = (TWFF_BMP, TWFF_TIFF);                                 }
      FileEnum      : TW_ENUMERATION
                    = (ItemType     : TWTY_INT16;
                       NumItems     : NUM_FILETYPES;
                       CurrentIndex : DEF_FILETYPE;
                       DefaultIndex : DEF_FILETYPE);


      { Units (Required type is Enumeration).                                  }
      NUM_UNITTYPES = 5;
      DEF_UNITTYPE  = 0;
      UnitType      : array[0..NUM_UNITTYPES-1] of TW_UINT16
                    = (TWUN_INCHES, TWUN_CENTIMETERS, TWUN_PICAS, TWUN_POINTS, TWUN_PIXELS);
      UnitEnum      : TW_ENUMERATION
                    = (ItemType     : TWTY_INT16;
                       NumItems     : NUM_UNITTYPES;
                       CurrentIndex : DEF_UNITTYPE;
                       DefaultIndex : DEF_UNITTYPE);


      { XferCount (Required type is OneValue).                                 }
      XferCount     : TW_UINT16 = 1;


      { Compression (Require type is OneValue).                                }
      Compression   : TW_UINT16 = TWCP_NONE;


      { Application changed resolution (xRes, yRes).                           }
      AppChangedRes : boolean = False;

      { xResolution (Required type is Range).                                  }
      xRes          : TW_RANGE
                    = (ItemType     : TWTY_FIX32;
                       MinValue     : TW_UINT32(1);
                       MaxValue     : TW_UINT32(32766);
                       StepSize     : TW_UINT32(1);
                       DefaultValue : TW_UINT32(300);
                       CurrentValue : TW_UINT32(300));


      { yResolution (Required type is Range).                                  }
      yRes          : TW_RANGE
                    = (ItemType     : TWTY_FIX32;
                       MinValue     : TW_UINT32(1);
                       MaxValue     : TW_UINT32(32766);
                       StepSize     : TW_UINT32(1);
                       DefaultValue : TW_UINT32(300);
                       CurrentValue : TW_UINT32(300));


      { Image Layout.                                                          }
      ImagePixFrame : TRect
                    = (Left   : 0;
                       Top    : 0;
                       Right  : 0;
                       Bottom : 0);

      ImageLayout   : TW_IMAGELAYOUT
                    = (Frame          : (Left   : (Whole : 0;
                                                   Frac  : 0);
                                         Top    : (Whole : 0;
                                                   Frac  : 0);
                                         Right  : (Whole : 0;
                                                   Frac  : 0);
                                         Bottom : (Whole : 0;
                                                   Frac  : 0));
                       DocumentNumber : TW_UINT32(1);
                       PageNumber     : TW_UINT32(1);
                       FrameNumber    : TW_UINT32(1));

{ Global variables                                                             }
{$IFDEF OLDSTYLE}
  var lpfnopenfileLB : TFarProc;
{$ENDIF}


implementation

uses WinProcs, SysUtils, Forms,
     {$IFDEF FALCON}
       ufalcon,
     {$ELSE}
       {$IFDEF IMASCAN}
         uimascan,
       {$ELSE}
         {$IFDEF TCITWAIN}
           utciMain,
         {$ELSE}
           uSrcTwn,
         {$ENDIF}
       {$ENDIF}
     {$ENDIF}
     twd_main, twd_caps,
     MsgBox;


{------------------------------------------------------------------------------}
{ MCM, Needed to handle large bitmaps.                                         }
{ This only applies to old 16 bit drivers.                                     }
{------------------------------------------------------------------------------}
{$IFNDEF WIN32}
  procedure AHIncr; far; external 'KERNEL' index 114;
{$ENDIF}

{------------------------------------------------------------------------------}
{ FUNCTION: Dispatch                                                           }
{                                                                              }
{ ARGS:    pTWMsg  ptr to the message struct passed to DS_ENTRY                }
{          pMapper ptr to a struct containizng ptr's to functions              }
{          DAT     DAT_xxxx that you are 'switching' on                        }
{                                                                              }
{ RETURNS: condition code                                                      }
{                                                                              }
{ NOTES:   This function will dispatch the correct function based on the       }
{          struct identifier.  Searches down through the given Map for the     }
{          matching DAT_xxxx type.                                             }
{------------------------------------------------------------------------------}

function Dispatch(pTWMsg  : PTWMessage;
                  pMapper : PMsgMap;
                  DAT     : TW_UINT16) : TW_UINT16;
var i       : integer;
    twRc    : TW_UINT16;
    Proc_Ok : boolean;
    s1, s2  : array[0..32] of char;
begin
  twRC    := TWRC_FAILURE;
  i       := 0;
  Proc_Ok := False;
  while Not(IsBadCodePtr(TFarProc(pMapper^[i].pFunc))) and Not(Proc_Ok)
  do begin
     if (pMapper^[i].ID = DAT)
     then begin
          Proc_Ok     := True;
          MsgFunction := TMsgFunction(pMapper^[i].pFunc);
          twRC        := MsgFunction(pTWMsg);
     end;
     inc(i);
  end;

  if Proc_Ok
  then Dispatch := twRC
  else begin
       DSStatus.ConditionCode := TWCC_BADPROTOCOL;
       Dispatch := TWRC_FAILURE;
  end;
end; { End Dispatch.                                                           }


{------------------------------------------------------------------------------}
{ FUNCTION: IdentityMsg                                                        }
{                                                                              }
{ ARGS:    pTWMsg  message struct, same as list of parms in DS_ENTRY call      }
{                                                                              }
{ RETURNS: twRC TWAIN status response                                          }
{                                                                              }
{ NOTES:   Handle the parse of: (from DAT_IDENTITY msg's)                      }
{              - MSG_OPENDS    open the indicated Source                       }
{              - MSG_CLOSEDS   close the indicated Source                      }
{              - MSG_GET       get identity struct of Source                   }
{              - error         error message                                   }
{                                                                              }
{------------------------------------------------------------------------------}

function IdentityMsg(pTWMsg : PTWMessage) : TW_UINT16;
var twRc : TW_UINT16;
    i    : integer;
begin
  LogMessage('IdentityMsg');
  twRc := TWRC_SUCCESS;
  case pTWMsg^.MSG of
  MSG_OPENDS  : OpenDS(pTWMsg);
  MSG_CLOSEDS : CloseDS;

  { Tell the SM who you are valid 3--7 states.                                 }
  MSG_GET     : pTW_IDENTITY(pTWMsg^.pData)^ := dsIdentity;
  else          begin
                  DSStatus.ConditionCode := TWCC_BADPROTOCOL;
                  twRc := TWRC_FAILURE;
                  DS_Error(E_IDENTITYMSG, DSStatus.ConditionCode);
                end;
  end;
  IdentityMsg := twRc;
end; { End IdentityMsg.                                                        }


{------------------------------------------------------------------------------}
{ FUNCTION: InterfaceMsg                                                       }
{                                                                              }
{ ARGS:    pTWMsg  message struct, same as list of parms in DS_ENTRY call      }
{                                                                              }
{ RETURNS: twRC TWAIN status response                                          }
{                                                                              }
{ NOTES:   Handle the parse of: (from DAT_USERINTERFACE msg's)                 }
{              - MSG_ENABLEDS  enable the open Source                          }
{              - MSG_DISABLEDS disable the open Source                         }
{              - errors        error message                                   }
{------------------------------------------------------------------------------}

function InterfaceMsg(pTWMsg : PTWMessage) : TW_UINT16;
var twRc : TW_UINT16;
begin
  LogMessage('InterfaceMsg');
  twRc := TWRC_SUCCESS;
  case pTWMsg^.MSG of
  MSG_ENABLEDS  : twRc := EnableDS(pTW_USERINTERFACE(pTWMsg^.pData));
  MSG_DISABLEDS : twRc := DisableDS(pTW_USERINTERFACE(pTWMsg^.pData));
  else            begin
                    DSStatus.ConditionCode := TWCC_BADPROTOCOL;
                    twRc := TWRC_FAILURE;
                    DS_Error(E_INTERFACEMSG, DSStatus.ConditionCode);
                  end;
  end;
  InterfaceMsg := twRc;
end; { End InterfaceMsg.                                                       }


{------------------------------------------------------------------------------}
{ FUNCTION: EnableDS                                                           }
{                                                                              }
{ ARGS:    pUI ptr to user interface struct                                    }
{                                                                              }
{ RETURNS: twRC TWAIN status response                                          }
{              - TWRC_SUCCESS                                                  }
{              - TWRC_FAILURE                                                  }
{                  - TWCC_SEQERROR                                             }
{                  - TWCC_LOWMEMORY                                            }
{                                                                              }
{ NOTES:                                                                       }
{ COMMENTS:    Valid states: 4, transitions to state 5 iff successful.         }
{              ShowUI bit 1=Source user interface; 0=app supplied              }
{              hParent field should contain a valid windows handle iff         }
{              you are implementing a modal user interface were the hparent    }
{              window will be disabled during access to the user IF dialog     }
{              via EnableWindow(hParent, False).  Else use hParent == NULL     }
{              for Modeless dialogs, see chapter 5.                            }
{------------------------------------------------------------------------------}

function EnableDS(pUI : pTW_USERINTERFACE) : TW_UINT16;
var twRC     : TW_UINT16;
begin
  LogMessage('EnableDS');
  twRC := TWRC_SUCCESS;
  if (DSState = STATE4)
  then begin { make this a global for later.                                   }
       hAppWnd := pUI^.hParent;  { need to know the app window's handle.       }
       FormSource := Nil;
       { Make a thunk for windows call back function                           }

       { If success, create a modeless dialog for modal use pUI^.hParent for   }
       { 3rd parm., see above.                                                 }

       {$IFDEF OLDSTYLE}
         lpfnopenfileLB := MakeProcInstance(@TWAIN_Interface, hDSInst);
         if (lpfnopenfileLB <> Nil)
         then begin
              hImageDlg := CreateWindow(TwainApp, 'TWAIN_Source',
                                        WS_OVERLAPPEDWINDOW,
                                        CW_USEDEFAULT,
                                        512,
                                        CW_USEDEFAULT,
                                        512,
                                        HWND_DESKTOP,
                                        LoadMenu(hDSInst, PChar(100)),
                                        hDSInst,
                                        Nil);
         end;

         { If either operation fails.                                          }
         if ((lpfnopenfileLB = Nil) or (hImageDlg = 0))
         then begin
              if (lpfnopenfileLB <> Nil)
              then FreeProcInstance(lpfnopenfileLB);
              lpfnopenfileLB := TFARPROC(Nil);
              DSStatus.ConditionCode := TWCC_LOWMEMORY;
              twRC := TWRC_FAILURE;
              DS_Error(E_ENABLEDS, DSStatus.ConditionCode);
         end
         { If both operations succeed.                                         }
         else begin
              { Show the window, IF user asks for it!                          }
              if pUI^.ShowUI
              then begin
                   LogMessage('Show DS window');
                   ShowWindow(hImageDlg, SW_SHOW);
                   UpdateWindow(hImageDlg);
              end
              else LogMessage('Hide DS window');
              DSState := STATE5;
         end;
       {$ELSE}
         FormSource := TFormSource.Create(Application);
         { If operation fails.                                                 }
         if (FormSource = Nil)
         then begin
              DSStatus.ConditionCode := TWCC_LOWMEMORY;
              twRC := TWRC_FAILURE;
              DS_Error(E_ENABLEDS, DSStatus.ConditionCode);
         end
         { If operations succeed.                                              }
         else begin
              hImageDlg := FormSource.Handle;
              { Show the window, IF user asks for it!                          }
              if (pUI^.ShowUI <> 0)
              then begin
                   FormSource.Show;
                   LogMessage('Show DS window');
              end
              else begin
                   LogMessage('Keep DS window hidden');

                   { Automatically get an image.                               }
                   FormSource.AutoOpenPic;

                   PostMessage(FormSource.Handle, 1999, 0, 0);
              end;
              DSState := STATE5;
         end;
       {$ENDIF}
  end
  else begin
       DSStatus.ConditionCode := TWCC_SEQERROR;
       twRC := TWRC_FAILURE;
       DS_Error(E_ENABLEDS, DSStatus.ConditionCode);
  end;
  EnableDS :=  twRC;
end; { End EnableDS.                                                           }


{------------------------------------------------------------------------------}
{ FUNCTION: DisableDS                                                          }
{                                                                              }
{ ARGS:    pUI     ptr to user interface structure                             }
{                                                                              }
{ RETURNS: TWRC_SUCCESS                                                        }
{          TWRC_FAILURE                                                        }
{              TWCC_SEQERROR                                                   }
{                                                                              }
{ NOTES:   Valid states: 5 through 7, transitions to 4 if successful.  The App }
{ sends down this call when the Source returns the TWMessage "MSG_CLOSEDSREQ". }
{                                                                              }
{ ShowUI bit 1=Source user interface; 0=app supplied                           }
{ hParent field should contain a valid windows handle                          }
{------------------------------------------------------------------------------}

function DisableDS(pUI : pTW_USERINTERFACE) : TW_UINT16;
var twRC : TW_UINT16;
begin
  LogMessage('DisableDS');
  twRC := TWRC_SUCCESS;
  case DSState of
  STATE5,  { Set hide UI if it was up.                                         }
  STATE6,
  STATE7 : begin
             {$IFDEF OLDSTYLE}
               if Not(pUI^.ShowUI)
               then ShowWindow(hImageDlg, SW_HIDE);
             {$ELSE}
               if (pUI^.ShowUI = 0) and Assigned(FormSource)
               then FormSource.Hide;
             {$ENDIF}

             { Transition to new state.                                        }
             DSState := STATE4;
           end;
      else begin
           DSStatus.ConditionCode := TWCC_SEQERROR;
           twRC := TWRC_FAILURE;
           DS_Error(E_DISABLEDS, DSStatus.ConditionCode);
      end;
  end;
  DisableDS := twRC;
end; { End DisableDS.                                                          }


{------------------------------------------------------------------------------}
{ FUNCTION: OpenDS                                                             }
{                                                                              }
{ ARGS:    pTWMsg  message struct, same as list of parms in DS_ENTRY call      }
{                                                                              }
{ RETURNS: status code                                                         }
{                                                                              }
{ NOTES:   This function will 'open' the Source, register it with the SM       }
{          Valid state: 3 transitions to state 4 on success.                   }
{ TODO: to support simultaneous connections to more than one app. the Source   }
{ should make a copy of it's states especially the identity structure of the   }
{ calling app.                                                                 }
{------------------------------------------------------------------------------}

function OpenDS(pTWMsg : PTWMessage) : TW_UINT16;
var twRC : TW_UINT16;
begin
  LogMessage('OpenDS');
  twRC := TWRC_SUCCESS;

  { The log is only good for one session of twaining to this source before     }
  { it is reset                                                                }
  OpenFile('c:\twsrc.log', Aof, OF_DELETE);

  if (DSState = STATE3)
  then begin
       { Only allow one connection at a time.  If AppIdentity.Id is set,       }
       { then there is already a connection.  If not, then remember who        }
       { we're connected to so we can send messages to it.                     }
       if (appIdentity.Id <> 0)
       then begin
            twRC := TWRC_FAILURE;
            DSStatus.ConditionCode := TWCC_MAXCONNECTIONS;
            DS_Error(E_OPENDS, DSStatus.ConditionCode);
       end
       else begin
            { This is to ensure that how we identify ourselves to the          }
            { SM is consistent with what it is keeping around.                 }
            appIdentity := pTW_IDENTITY(pTWMsg^.pSrc)^;
            dsIdentity  := pTW_IDENTITY(pTWMsg^.pData)^;
            {$IFDEF DEBUG}
              ShowTW_ID(hImageDlg, appIdentity, 'App ID (Source)');
            {$ENDIF}
            DSState := STATE4;
       end;
  end
  else begin
       twRC := TWRC_FAILURE;
       DSStatus.ConditionCode := TWCC_SEQERROR;
       DS_Error(E_OPENDS, DSStatus.ConditionCode);
  end;
  OpenDS := twRC;
end; { End OpenDS.                                                             }


{------------------------------------------------------------------------------}
{ FUNCTION: CloseDS                                                            }
{                                                                              }
{ ARGS:    none                                                                }
{                                                                              }
{ RETURNS: status code                                                         }
{                                                                              }
{ NOTES:   This function will close the Source.                                }
{------------------------------------------------------------------------------}

function CloseDS : TW_UINT16;
var twRC : TW_UINT16;
begin
  LogMessage('CloseDS');
  twRC := TWRC_SUCCESS;
  case DSState of
  STATE4,
  STATE5,
  STATE6,
  STATE7 : begin { Clear out current connection.                               }
             appIdentity.Id := 0;

             {$IFDEF OLDSTYLE}
               { Free up the window.                                           }
               if (hImageDlg <> 0)
               then DestroyWindow(hImageDlg);

               { Remove the proc instance.                                     }
               if (lpfnopenfileLB <> Nil)
               then begin
                    FreeProcInstance(lpfnopenfileLB);
                    lpfnopenfileLB := TFARPROC(Nil);
               end;
             {$ELSE}
             { This does not work !!!!!
                try
                  FormSource.Close;
                finally
                  FormSource := Nil;
                end;
             }
             {$ENDIF}

             { Set the DS state.                                               }
             DSState := STATE3;
           end;
  else     begin
             twRC := TWRC_FAILURE;
             DSStatus.ConditionCode := TWCC_SEQERROR;
             DS_Error(E_CLOSEDS, DSStatus.ConditionCode);
           end;
  end;
  CloseDS := twRC;
end; { End CloseDS.                                                            }


{------------------------------------------------------------------------------}
{ Sub functions for CapabilityMsg.                                             }
{------------------------------------------------------------------------------}

function GetSupportCap(pTWMsg : PTWMessage) : TW_UINT16;
begin
  Result := TWRC_SUCCESS;
  BuildUpArrayType(pTW_CAPABILITY(pTWMsg^.pData),
                   @SupportedCaps,
                   @SupportedTypes,
                   pTWMsg^.Msg);
end; { End GetSupportCap.                                                      }


function SetXferMechCap(pTWMsg : PTWMessage; var twRc : TW_UINT16) : TW_UINT16;
var pData : pTW_CAPABILITY;
    Value : TW_UINT16;
    i     : integer;
begin
  Result := TWRC_SUCCESS;
  pData  := pTWMsg^.pData;
  case pData^.ConType of
  { Search thru XferType array and set CurrentIndex if value of Temp           }
  { is found else, set closest and ask user to check status. TODO.             }
  TWON_ONEVALUE    : begin
                       ExtractOneValue(pData, @Value);
                       XferEnum.CurrentIndex := 0;
                       for i := 0 to (XferEnum.NumItems - 1)
                       do if (Value = XferType[i])
                          then XferEnum.CurrentIndex := i;
                     end;
  { Now that you have the value(s) what do you do with them ??                 }
  { You should use them to set up a filter on this capability                  }
  { to restrict the choices for this CAP BEFORE showing the                    }
  { choices to the end user.                                                   }
  { Example: Source (scanner) supports 50 100 150 200 250 300 DPI              }
  { the App ONLY supports 100 200 300                                          }
  { therefore when Sources allows the user to set DPI, ONLY 100                }
  { 200 300 should be shown as available options.                              }
  { App, be sure to free the container                                         }
  TWON_ENUMERATION : begin
                       ExtractEnumerationIndex(pData, XferEnum.CurrentIndex);
                       ExtractEnumerationValue(pData, @Value,
                                               XferEnum.CurrentIndex);
                     end;
  else begin { Error, non-supported type.                                      }
       Result := TWCC_BADCAP;
       twRC   := TWRC_FAILURE;
  end;
  end;
end; { End SetXferMechCap.                                                     }


function GetXferMechCap(pTWMsg : PTWMessage) : TW_UINT16;
begin
  Result := TWRC_SUCCESS;
  BuildUpEnumerationType(pTW_CAPABILITY(pTWMsg^.pData),
                         @XferEnum,
                         @XferType,
                         pTWMsg^.Msg);
end; { End GetXferMechCap.                                                     }


function SetXferCountCap(pTWMsg : PTWMessage; var twRc : TW_UINT16) : TW_UINT16;
begin
  Result := TWRC_SUCCESS;
  case pTW_CAPABILITY(pTWMsg^.pData)^.ConType of
  TWON_ONEVALUE    : begin
                       ExtractOneValue(pTW_CAPABILITY(pTWMsg^.pData),
                                       @XferCount);
                       XferCount := 1; { Source will only transfer one image.  }
                     end;
  else begin { Error, I don't support that container type                      }
       Result := TWCC_BADCAP;
       twRC   := TWRC_FAILURE;
  end;
  end;
end; { End SetXferCountCap.                                                    }


function GetXferCountCap(pTWMsg : PTWMessage) : TW_UINT16;
begin
  Result := TWRC_SUCCESS;
  BuildUpOneValue(pTW_CAPABILITY(pTWMsg^.pData),
                  TWTY_UINT16,
                  @XferCount,
                  pTWMsg^.Msg);
end; { End GetXferCountCap.                                                    }


function SetFileFormatCap(pTWMsg : PTWMessage; var twRc : TW_UINT16) : TW_UINT16;
var pData : pTW_CAPABILITY;
    Value : TW_UINT16;
    i     : integer;
begin
  Result := TWRC_SUCCESS;
  pData  := pTWMsg^.pData;
  case pData^.ConType of
  { Search thru UnitType array and set CurrentIndex iff value of Temp          }
  { is found else, set closest and ask user to check status. TODO.             }
  TWON_ONEVALUE    : begin
                       ExtractOneValue(pData, @Value);
                       FileEnum.CurrentIndex := 0;
                       for i := 0 to (FileEnum.NumItems - 1)
                       do if (Value = FileType[i])
                          then FileEnum.CurrentIndex := i;
                     end;
  TWON_ENUMERATION : begin
                       ExtractEnumerationIndex(pData, FileEnum.CurrentIndex);
                       ExtractEnumerationValue(pData, @Value,
                                               FileEnum.CurrentIndex);
                     end;
  else begin { Error, non-supported type.                                      }
       Result := TWCC_BADCAP;
       twRC   := TWRC_FAILURE;
  end;
  end;
end; { End SetFileFormatCap.                                                   }


function GetFileFormatCap(pTWMsg : PTWMessage) : TW_UINT16;
begin
  Result := TWRC_SUCCESS;
  BuildUpEnumerationType(pTW_CAPABILITY(pTWMsg^.pData),
                         @FileEnum,
                         @FileType,
                         pTWMsg^.Msg);
end; { End GetFileFormatCap.                                                   }


function SetUnitsCap(pTWMsg : PTWMessage; var twRc : TW_UINT16) : TW_UINT16;
var pData : pTW_CAPABILITY;
    Value : TW_UINT16;
    i     : integer;
begin
  Result := TWRC_SUCCESS;
  pData  := pTWMsg^.pData;
  case pData^.ConType of
  { Search thru UnitType array and set CurrentIndex iff value of Temp          }
  { is found else, set closest and ask user to check status. TODO.             }
  TWON_ONEVALUE    : begin
                       ExtractOneValue(pData, @Value);
                       UnitEnum.CurrentIndex := 0;
                       for i := 0 to (UnitEnum.NumItems - 1)
                       do if (Value = UnitType[i])
                          then UnitEnum.CurrentIndex := i;
                     end;
  TWON_ENUMERATION : begin
                       ExtractEnumerationIndex(pData, UnitEnum.CurrentIndex);
                       ExtractEnumerationValue(pData, @Value,
                                               UnitEnum.CurrentIndex);
                     end;
  else begin { Error, non-supported type.                                      }
       Result := TWCC_BADCAP;
       twRC   := TWRC_FAILURE;
  end;
  end;
end; { End SetUnitsCap.                                                        }


function GetUnitsCap(pTWMsg : PTWMessage) : TW_UINT16;
begin
  Result := TWRC_SUCCESS;
  UnitEnum.NumItems := 2;
  BuildUpEnumerationType(pTW_CAPABILITY(pTWMsg^.pData),
                         @UnitEnum,
                         @UnitType,
                         pTWMsg^.Msg);
end; { End GetUnitsCap.                                                        }


function SetPixelTypeCap(pTWMsg : PTWMessage; var twRc : TW_UINT16) : TW_UINT16;
var pData : pTW_CAPABILITY;
    Value : TW_UINT16;
    i     : integer;
begin
  Result := TWRC_SUCCESS;
  pData  := pTWMsg^.pData;
  case pData^.ConType of
  { TODO change value.                                                         }
  TWON_ONEVALUE    : begin
                       ExtractOneValue(pTW_CAPABILITY(pTWMsg^.pData), @Value);
                       PixelEnum.CurrentIndex := 0;
                       for i := 0 to (PixelEnum.NumItems - 1)
                       do if (Value = PixelType[i])
                          then PixelEnum.CurrentIndex := i;
                       AppChangedPix := True;
                     end;
  TWON_ENUMERATION : begin
                       ExtractEnumerationIndex(pData, PixelEnum.CurrentIndex);
                       ExtractEnumerationValue(pData, @Value,
                                               PixelEnum.CurrentIndex);
                       AppChangedPix := True;
                     end;
  else begin { Error, non-supported type.                                      }
       Result := TWCC_BADCAP;
       twRC   := TWRC_FAILURE;
  end;
  end;
end; { End SetPixelTypeCap.                                                    }


function GetPixelTypeCap(pTWMsg : PTWMessage) : TW_UINT16;
var Value : TW_UINT32;
    i     : integer;
begin
  Result := TWRC_SUCCESS;
  if Assigned(FormSource)
  then begin
       Value := TWPT_RGB;
       if Assigned(FormSource)
       then begin
            if (FormSource.ImageStream.Size = 0)
            then FormSource.CopyBmpToStream;
            if (FormSource.pBmpInfo <> Nil)
            then begin
                 Value := FormSource.pBmpInfo^.bmiHeader.biBitCount;
                 case Value of
                 1  : Value := TWPT_BW;
                 24,
                 32 : Value := TWPT_RGB;
                 else begin
                      if IsPaletteGray
                      then Value := TWPT_GRAY
                      else Value := TWPT_PALETTE;
                 end;
                 end;
            end;
       end;
       PixelEnum.CurrentIndex := 0;
       for i := 0 to (PixelEnum.NumItems - 1)
       do if (Value = PixelType[i])
          then PixelEnum.CurrentIndex := i;
  end;
  BuildUpEnumerationType(pTW_CAPABILITY(pTWMsg^.pData),
                         @PixelEnum,
                         @PixelType,
                         pTWMsg^.Msg);
end; { End GetPixelTypeCap.                                                    }


function SetCompressCap(pTWMsg : PTWMessage; var twRc : TW_UINT16) : TW_UINT16;
var pData        : pTW_CAPABILITY;
    pEnumeration : pTW_ENUMERATION;
begin
  Result := TWRC_SUCCESS;
  pData := pTWMsg^.pData;
  case pData^.ConType of
  TWON_ONEVALUE    : begin
                       ExtractOneValue(pTW_CAPABILITY(pTWMsg^.pData), @Compression);
                       { Don't allow compression !                             }
                       Compression := TWCP_NONE;
                     end;
  TWON_ENUMERATION : begin
                       pEnumeration := pTW_ENUMERATION(GlobalLock(pData^.hContainer));
                       ExtractEnumerationValue(pData, @Compression,
                                               pEnumeration^.CurrentIndex);
                       GlobalUnlock(pData^.hContainer);

                       Compression := TWCP_NONE;
                     end;
  else begin { Error, I don't support that container type                      }
       Result := TWCC_BADCAP;
       twRC   := TWRC_FAILURE;
  end;
  end;
end; { End SetCompressCap.                                                     }


function GetCompressCap(pTWMsg : PTWMessage) : TW_UINT16;
begin
  Result := TWRC_SUCCESS;
  BuildUpOneValue(pTW_CAPABILITY(pTWMsg^.pData),
                  TWTY_UINT16,
                  @Compression,
                  pTWMsg^.Msg);
end; { End GetCompressCap.                                                     }


function SetXresCap(pTWMsg : PTWMessage; var twRc : TW_UINT16) : TW_UINT16;
var pData        : pTW_CAPABILITY;
    pRangeValue  : pTW_RANGE;
    pEnumeration : pTW_ENUMERATION;
    NewValue     : TW_UINT32;
begin
  Result := TWRC_SUCCESS;
  pData := pTWMsg^.pData;
  AppChangedRes := True;
  case pData^.ConType of
  TWON_ONEVALUE    : ExtractOneValue(pTW_CAPABILITY(pTWMsg^.pData), @xRes.CurrentValue);
  TWON_ENUMERATION : begin
                       pEnumeration := pTW_ENUMERATION(GlobalLock(pData^.hContainer));
                       ExtractEnumerationValue(pData, @NewValue,
                                               pEnumeration^.CurrentIndex);
                       GlobalUnlock(pData^.hContainer);

                       if (NewValue < xRes.MinValue)
                       then xRes.CurrentValue := xRes.MinValue
                       else if (NewValue > xRes.MaxValue)
                            then xRes.CurrentValue := xRes.MaxValue
                            else xRes.CurrentValue := NewValue;
                     end;
  TWON_RANGE       : begin
		       pRangeValue := pTW_RANGE(GlobalLock(pData^.hContainer));
                       { Get default value.                                    }
                       if (pRangeValue^.DefaultValue < xRes.MinValue)
                       then xRes.DefaultValue := xRes.MinValue
                       else if (pRangeValue^.DefaultValue > xRes.MaxValue)
                            then xRes.DefaultValue := xRes.MaxValue
                            else xRes.DefaultValue := pRangeValue^.DefaultValue;

                       { Get current value.                                    }
                       if (pRangeValue^.CurrentValue < xRes.MinValue)
                       then xRes.CurrentValue := xRes.MinValue
                       else if (pRangeValue^.CurrentValue > xRes.MaxValue)
                            then xRes.CurrentValue := xRes.MaxValue
                            else xRes.CurrentValue := pRangeValue^.CurrentValue;
		       GlobalUnlock(pData^.hContainer);
                     end;
  else begin { Error, I don't support that container type                      }
       Result := TWCC_BADCAP;
       twRC   := TWRC_FAILURE;
  end;
  end;
end; { End SetXresCap.                                                         }


function GetXresCap(pTWMsg : PTWMessage) : TW_UINT16;
var DefValue   : TW_UINT32;
begin
  Result := TWRC_SUCCESS;
  DefValue := 300; { Default reply                                             }
  if Assigned(FormSource)
  then begin
       if (FormSource.ImageStream.Size = 0)
       then FormSource.CopyBmpToStream;
       if (FormSource.pBmpInfo <> Nil)
       then DefValue := FormSource.pBmpInfo^.bmiHeader.biXPelsPerMeter;
  end;
  BuildUpRangeType(pTW_CAPABILITY(pTWMsg^.pData),
                   @xRes,
                   pTWMsg^.Msg);
end; { End GetXresCap.                                                         }


function SetYresCap(pTWMsg : PTWMessage; var twRc : TW_UINT16) : TW_UINT16;
var pData        : pTW_CAPABILITY;
    pRangeValue  : pTW_RANGE;
    pEnumeration : pTW_ENUMERATION;
    NewValue     : TW_UINT32;
begin
  Result := TWRC_SUCCESS;
  pData := pTWMsg^.pData;
  AppChangedRes := True;
  case pData^.ConType of
  TWON_ONEVALUE    : ExtractOneValue(pTW_CAPABILITY(pTWMsg^.pData), @yRes.CurrentValue);
  TWON_ENUMERATION : begin
                       pEnumeration := pTW_ENUMERATION(GlobalLock(pData^.hContainer));
                       ExtractEnumerationValue(pData, @NewValue,
                                               pEnumeration^.CurrentIndex);
                       GlobalUnlock(pData^.hContainer);

                       if (NewValue < yRes.MinValue)
                       then yRes.CurrentValue := yRes.MinValue
                       else if (NewValue > yRes.MaxValue)
                            then yRes.CurrentValue := yRes.MaxValue
                            else yRes.CurrentValue := NewValue;
                     end;
  TWON_RANGE       : begin
		       pRangeValue := pTW_RANGE(GlobalLock(pData^.hContainer));
                       { Get default value.                                    }
                       if (pRangeValue^.DefaultValue < yRes.MinValue)
                       then yRes.DefaultValue := yRes.MinValue
                       else if (pRangeValue^.DefaultValue > yRes.MaxValue)
                            then yRes.DefaultValue := yRes.MaxValue
                            else yRes.DefaultValue := pRangeValue^.DefaultValue;

                       { Get current value.                                    }
                       if (pRangeValue^.CurrentValue < yRes.MinValue)
                       then yRes.CurrentValue := yRes.MinValue
                       else if (pRangeValue^.CurrentValue  > yRes.MaxValue)
                            then yRes.CurrentValue := yRes.MaxValue
                            else yRes.CurrentValue := pRangeValue^.CurrentValue;
		       GlobalUnlock(pData^.hContainer);
                     end;
  else begin { Error, I don't support that container type                      }
       Result := TWCC_BADCAP;
       twRC   := TWRC_FAILURE;
  end;
  end;
end; { End SetYresCap.                                                         }


function GetYresCap(pTWMsg : PTWMessage) : TW_UINT16;
var Value      : TW_UINT32;
begin
  Result := TWRC_SUCCESS;
  Value := 300;
  if Assigned(FormSource)
  then begin
       if (FormSource.ImageStream.Size = 0)
       then FormSource.CopyBmpToStream;
       if (FormSource.pBmpInfo <> Nil)
       then Value := FormSource.pBmpInfo^.bmiHeader.biYPelsPerMeter;
  end;
  BuildUpRangeType(pTW_CAPABILITY(pTWMsg^.pData),
                   @yRes,
                   pTWMsg^.Msg);
end; { End GetYresCap.                                                         }


{------------------------------------------------------------------------------}
{ FUNCTION: CapabilityMsg                                                      }
{                                                                              }
{ ARGS:    pTWMsg  message struct, same as list of parms in DS_ENTRY call      }
{                                                                              }
{ RETURNS: status code                                                         }
{                                                                              }
{ NOTES: for the MSG_SET side of Source                                        }
{ The application will create the container and destroy the container.         }
{ The Source will extract the value from the container 1st and use this value  }
{ to modify it's local value for the CAP.                                      }
{                                                                              }
{ The Source needs to be able to extract values from all of the different      }
{ container types and load as many as possible into it's local containers      }
{ for the particular capabilty.                                                }
{                                                                              }
{ NOTES: for the MSG_GET side of Source                                        }
{ The Source will create the container and the App will destroy container.     }
{ The App will extract the value from the container 1st and use this value     }
{ to modify it's local state or value for this CAP.                            }
{                                                                              }
{ The App extracts values from all different container types and also          }
{ matches up these cap's when appropriate to find the value or values which    }
{ best suit the current mode of the application.                               }
{                                                                              }
{ NOTES: for both App and Source in general:                                   }
{ You don't know what type of container you will be getting for any            }
{ particular cap until you get it.                                             }
{------------------------------------------------------------------------------}

function CapabilityMsg(pTWMsg : PTWMessage) : TW_UINT16;
var twRC       : TW_UINT16;
    Number     : array[0..8] of char;
    TwackerIni : array[0..MAX_TWPATH] of char;
    Capability : array[0..80] of char;
    NewValue   : TW_UINT32;
    Fix32Value : TW_FIX32;
    Ret        : TW_UINT16;
    i          : integer;
    GetSetStr  : string;
    TmpStr     : string;
begin
  LogMessage('CapabilityMsg');
  NewValue := 5;
  twRC := TWRC_SUCCESS;
  { Since the source control file only manages capabilities, it can be checked }
  { here.  If this file is ever used for anything else, it would be good to    }
  { put these calls in a more general place.  DllMain is not a good idea though}
  { because of the re-entrant nature of a TWAIN source.                        }
  { always try to set the source control in case it is changed on the fly      }
  if Not(SetSourceControl)
  then twRC := TWRC_SUCCESS
  else begin
       { Convert the cap number to the string representation, retreived from   }
       { d:\WINDOWS\TWACKER.INI                                                }
       GetWindowsDirectory(TwackerIni, SizeOf(TwackerIni));
       lstrcat(TwackerIni, '\twacker.ini');
       Str(pTW_CAPABILITY(pTWMsg^.pData)^.Cap, Number);
       GetPrivateProfileString('CAP NUMBERS',
                               Number,
                               'NONE',
                               Capability,
                               SizeOf(Capability),
                               TwackerIni);
       {$IFDEF DEBUG}
         case pTWMsg^.MSG of
         MSG_GET        : GetSetStr := ' \ MSG_GET';
         MSG_GETCURRENT : GetSetStr := ' \ MSG_GETCURRENT';
         MSG_GETDEFAULT : GetSetStr := ' \ MSG_GETDEFAULT';
         MSG_SET        : GetSetStr := ' \ MSG_SET';
         MSG_RESET      : GetSetStr := ' \ MSG_RESET';
         else GetSetStr := '- ERROR, BAD PROTECOL -';
         end;
         TmpStr := StrPas(Capability) + GetSetStr + ' #' + StrPas(Number);
         LogMessage(PChar(TmpStr));

         if (pTWMsg^.MSG = MSG_SET)
         then begin
              case pTW_CAPABILITY(pTWMsg^.pData)^.ConType of
              TWON_ARRAY       : GetSetStr := 'TWON_ARRAY';
              TWON_ENUMERATION : GetSetStr := 'TWON_ENUMERATION';
              TWON_ONEVALUE    : GetSetStr := 'TWON_ONEVALUE';
              TWON_RANGE       : GetSetStr := 'TWON_RANGE';
              else GetSetStr := '- ERROR -';
              end;
              GetSetStr := 'ConType: ' + GetSetStr;
              LogMessage(PChar(GetSetStr));
         end;
       {$ENDIF}


       if (lstrcmp(Capability, 'NONE') = 0)
       then begin
            DSStatus.ConditionCode := TWCC_BADCAP;
            twRc := TWRC_FAILURE;
       end
       else begin
            case pTWMsg^.MSG of
            MSG_GET,
            MSG_GETCURRENT,    { I would assume these should be handled        }
            MSG_GETDEFAULT : begin
                               { MCM,                                          }
                               { If you would like to control the return values}
                               { to the app add a case.                        }
                               Ret  := TWRC_SUCCESS;
                               case pTW_CAPABILITY(pTWMsg^.pData)^.Cap of
                               { Get supported capabilities.                   }
                               CAP_SUPPORTEDCAPS : Ret := GetSupportCap(pTWMsg);

                               { Required support return the CAP_ in an        }
                               { Enumeration container.                        }
                               { ICAP_XFERMECH values (SX_ means Setup XFer)   }
                               { constants: DCSX_NATIVE, DCSX_FILE, DCSX_MEMORY}
                               ICAP_XFERMECH    : Ret := GetXferMechCap(pTWMsg);

                               ICAP_IMAGEFILEFORMAT : Ret := GetFileFormatCap(pTWMsg);

                               { Required support return the CAP_ in an        }
                               { Enumeration container.                        }
                               { ICAP_UNITS values (UN_ means UNits)           }
                               { constants: TWUN_INCHES, TWUN_CENTIMETERS,     }
                               {            TWUN_PICAS,  TWUN_POINTS,          }
                               {            TWUN_TWIPS,  TWUN_NONE             }
                               { MCM, I'll limit unit type to Inch and cm.     }
                               ICAP_UNITS       : Ret := GetUnitsCap(pTWMsg);

                               { Required support return the CAP_ in an        }
                               { Enumeration container.                        }
                               { constants: TWPT_BW,  TWPT_GRAY,               }
                               {            TWPT_RGB, TWPT_PALETTE             }
                               ICAP_PIXELTYPE   : Ret := GetPixelTypeCap(pTWMsg);

                               { Required support return the CAP_ in a         }
                               { OneValue container.                           }
                               CAP_XFERCOUNT    : Ret := GetXferCountCap(pTWMsg);
                               { constants: TWCP_NONE, TWCP_PACKBITS...TWCP_LZW}

                               { I have chosen to return the CAP_ in a         }
                               { OneValue container type.                      }
                               ICAP_COMPRESSION : Ret := GetCompressCap(pTWMsg);

                               ICAP_XRESOLUTION : Ret := GetXresCap(pTWMsg);
                               ICAP_YRESOLUTION : Ret := GetYresCap(pTWMsg);
(*                               
                               CAP_AUTOFEED : begin
                                                Ret := 0;
                                                {
                                                GetPrivateProfileInt('CAP_XFERCOUNT', 'Current', 0, SourceControl);
                                                WritePrivateProfileString('ADF','PendingImages', '0', SourceControl);
                                                }
                                              end;
*)
                               else begin { MCM, Get default message handling. }
                                    { If you do not want default message
                                    { handling insert the two lines below, and }
                                    { remark out the thried.                   }
                                    { Ret  := TWCC_BADCAP;                     }
                                    { twRC := TWRC_FAILURE;                    }
                                    Ret := SetValues(pTW_CAPABILITY(pTWMsg^.pData),
                                                     pTWMsg^.Msg,
                                                     Capability);
                               end;
                               end;
                               if (Ret <> TWCC_SUCCESS)
                               then twRc := TWRC_FAILURE;
                               DSStatus.ConditionCode := Ret;
                             end;
            MSG_SET        : begin
                               { MCM,                                          }
                               { If you would like to control the return values}
                               { to the app add a case.                        }
                               Ret  := TWRC_SUCCESS;
                               case pTW_CAPABILITY(pTWMsg^.pData)^.Cap of
                               { Determine container type, then call           }
                               { appropriate extract routine.                  }
                               ICAP_XFERMECH    : Ret := SetXferMechCap(pTWMsg, twRc);

                               ICAP_IMAGEFILEFORMAT : Ret := SetFileFormatCap(pTWMsg, twRc);
                               { Determine container type, then call           }
                               { appropriate extract routine.                  }
                               ICAP_UNITS       : Ret := SetUnitsCap(pTWMsg, twRc);

                               { Determine container type, then call           }
                               { appropriate extract routine.                  }
                               ICAP_PIXELTYPE   : Ret := SetPixelTypeCap(pTWMsg, twRc);

                               { Required container type is ONLY OneValue.     }
                               CAP_XFERCOUNT    : Ret := SetXferCountCap(pTWMsg, twRc);

                               ICAP_COMPRESSION : Ret := SetCompressCap(pTWMsg, twRc);
                               ICAP_XRESOLUTION : Ret := SetXresCap(pTWMsg, twRc);
                               ICAP_YRESOLUTION : Ret := SetYresCap(pTWMsg, twRc);
                               { FeederEnabled special case see page 9-9 twain }
                               { manual.                                       }
                               { if (pTW_CAPABILITY(pTWMsg^.pData)^.Cap =      }
                               {     CAP_FEEDERENABLED)                        }
                               { then begin                                    }
                               {      Ret  := TWCC_BADVALUE;                   }
                               {      twRc := TWRC_FAILURE;                    }
                               { end;                                          }
                               else begin
                                    { MCM, Default message handling.           }
                                    Ret := SetNewCap(pTWMsg^.pData, Capability);
                                    if (Ret = TWRC_SUCCESS)
                                    then SetValues(pTW_CAPABILITY(pTWMsg^.pData),
                                                   pTWMsg^.MSG,
                                                   Capability)
                                    else twRc := TWRC_FAILURE;
                                    { If you do not want default message       }
                                    { handling insert the two lines below      }
                                    { Ret  := TWCC_BADCAP;                     }
                                    { twRC := TWRC_FAILURE;                    }
                               end;
                               end;

                               DSStatus.ConditionCode := Ret;
                             end;
            MSG_RESET      : begin
                               DSStatus.ConditionCode := ResetValues(Capability);
                               if (DSStatus.ConditionCode <> TWCC_BADPROTOCOL)
                               then SetValues(pTW_CAPABILITY(pTWMsg^.pData),
                                              pTWMsg^.MSG,
                                              Capability)
                               else twRc := TWRC_FAILURE;
                             end;
            else             begin
                               DSStatus.ConditionCode := TWCC_BADPROTOCOL;
                               twRC := TWRC_FAILURE;
                               DS_Error(E_CAPABILITYMSG, DSStatus.ConditionCode);
                             end;
            end; { End of TWMsg.                                               }
       end;
       {$IFDEF DEBUG}
         if (pTWMsg^.MSG <> MSG_SET) and
            (pTWMsg^.MSG <> MSG_RESET)
         then begin
              case pTW_CAPABILITY(pTWMsg^.pData)^.ConType of
              TWON_ARRAY       : GetSetStr := 'TWON_ARRAY';
              TWON_ENUMERATION : GetSetStr := 'TWON_ENUMERATION';
              TWON_ONEVALUE    : GetSetStr := 'TWON_ONEVALUE';
              TWON_RANGE       : GetSetStr := 'TWON_RANGE';
              else GetSetStr := '- ERROR -';
              end;
              GetSetStr := 'ConType: ' + GetSetStr;
              LogMessage(PChar(GetSetStr));
         end;
       {$ENDIF}
  end;
  CapabilityMsg := twRC;
end; { End CapabilityMsg.                                                      }


{------------------------------------------------------------------------------}
{ FUNCTION: StatusMsg                                                          }
{                                                                              }
{ ARGS:    pTWMsg  message struct, same as list of parms in DS_ENTRY call      }
{                                                                              }
{ RETURNS: status code                                                         }
{                                                                              }
{ NOTES:   Valid states 3 through 7.  Source fills in all fields of pDSMStatus }
{ and clears it's condition code.                                              }
{------------------------------------------------------------------------------}

function StatusMsg(pTWMsg : PTWMessage) : TW_UINT16;
var twRC : TW_UINT16;
begin
  twRC := TWRC_SUCCESS;
  case pTWMsg^.MSG of
  MSG_GET : begin
              pTW_STATUS(pTWMsg^.pData)^ := DSStatus; { copy all fields        }
              DSStatus.ConditionCode := TWCC_SUCCESS; { clear condition code   }
            end;
  else      begin { Invalid triplet.                                           }
              DSStatus.ConditionCode := TWCC_BADPROTOCOL;
              twRC := TWRC_FAILURE;
              DS_Error(E_STATUSMSG, DSStatus.ConditionCode);
            end;
  end;
  StatusMsg := twRC;
end; { End StatusMsg.                                                          }


{------------------------------------------------------------------------------}
{ FUNCTION: Msg                                                                }
{                                                                              }
{ ARGS:    pTWMsg  message struct, same as list of parms in DS_ENTRY call      }
{                                                                              }
{ RETURNS: TWRC_SUCCESS                                                        }
{          TWRC_FAILURE                                                        }
{              TWCC_SEQERROR                                                   }
{              TWCC_BADPROTOCOL        // debug ONLY                           }
{                                                                              }
{ NOTES:   Valid states: 4 -- 7.                                               }
{------------------------------------------------------------------------------}

function PendingXferMsg(pTWMsg : PTWMessage) : TW_UINT16;
var twRC          : TW_UINT16;
    Buffer        : array[0..7] of char;
    PendingImages : TW_INT16;
begin
  LogMessage('PendingXferMsg');
  twRC := TWRC_SUCCESS;
  PendingImages := 0;
  case pTWMsg^.MSG of
  MSG_GET       : begin { Valid states 4 -- 7.                                 } 
                    case DSState of
                    { single image supported, always 0.                        }
                    STATE4,
                    STATE5,
                    STATE6,
                    STATE7 : pTW_PENDINGXFERS(pTWMsg^.pData)^.Count :=
                             GetPrivateProfileInt('ADF',
                                                  'PendingImages',
                                                  0,
                                                  SourceControl);
                    else     begin { Fail get due to  seqerror.                }
                               DSStatus.ConditionCode := TWCC_SEQERROR;
                               twRC := TWRC_FAILURE;
                               DS_Error(E_PENDINGXFERMSG, DSStatus.ConditionCode);
                             end;
                    end;
                  end;
    MSG_ENDXFER : begin { clears ONLY current transfer, valid 6 or 7           }
                    { pending images needs to be set somewhere else --         }
                    { this is a patch                                          }
                    PendingImages := 0;
                    {GetPrivateProfileInt('ADF',
                                          'PendingImages',
                                          0,
                                          SourceControl);
                    }
                    dec(PendingImages);
                    if (PendingImages < 0)
                    then PendingImages := 0;

                    pTW_PENDINGXFERS(pTWMsg^.pData)^.Count := TW_UINT16(PendingImages);
                    Str(PendingImages, buffer);
                    WritePrivateProfileString('ADF',
                                              'PendingImages',
                                              buffer,
                                              SourceControl);
                  end;
    MSG_RESET   : { Clears ALL transfers, valid 6 or 7                         }
                  if ((DSState = STATE6) or (DSState = STATE7))
                  then begin
                       { Always transition to state 5 since I only support     }
                       { single image.                                         }
                       DSState := STATE5;
                       pTW_PENDINGXFERS(pTWMsg^.pData)^.Count := 0;
                       WritePrivateProfileString('ADF',
                                                 'PendingImages',
                                                 '0',
                                                 SourceControl);
                  end
                  else begin
                       DSStatus.ConditionCode := TWCC_SEQERROR;
                       twRC := TWRC_FAILURE;
                       DS_Error(E_PENDINGXFERMSG, DSStatus.ConditionCode);
                  end;
  else            begin
                    DSStatus.ConditionCode := TWCC_BADPROTOCOL;
                    twRC := TWRC_FAILURE;
                    DS_Error(E_PENDINGXFERMSG, DSStatus.ConditionCode);
                  end;
  end;
  PendingXferMsg := twRC;
end; { End PendingXferMsg.                                                     }


{------------------------------------------------------------------------------}
{ FUNCTION: SetupMemXferMsg                                                    }
{                                                                              }
{ ARGS:    pTWMsg  message struct, same as list of parms in DS_ENTRY call      }
{                                                                              }
{ RETURNS: TWRC_SUCCESS                                                        }
{          TWRC_FAILURE                                                        }
{              TWCC_SEQERROR                                                   }
{              TWCC_BADPROTOCOL        // debug ONLY                           }
{                                                                              }
{ NOTES:   Valid states: 4 -- 6.                                               }
{ Source fills in the following:                                               }
{------------------------------------------------------------------------------}

function SetupMemXferMsg(pTWMsg : PTWMessage) : TW_UINT16;
var twRC         : TW_UINT16;
    Pref         : integer;
    NumColors    : integer;
    LongWidth    : integer;
    SetupMemXfer : pTW_SETUPMEMXFER;
begin
  LogMessage('SetupMemXferMsg');
  XferDone  := False;
  XferFirst := True;
  BytesOut  := 0;

  // for DEBUG only.
  XferCount := 0;

  twRC := TWRC_SUCCESS;
  case DSState of
  STATE4,
  STATE5,
  STATE6 : case pTWMsg^.MSG of
           MSG_GET : begin
                       SetupMemXfer := pTW_SETUPMEMXFER(pTWMsg^.pData);
                       MinBufSize := 0;
                       MaxBufSize := 0;

                       if Assigned(FormSource)
                       then begin
                            if (FormSource.ImageStream.Size = 0)
                            then FormSource.CopyBmpToStream;
                            if (FormSource.ImageStream.Size > 0)
                            then begin
                                 with FormSource.pBmpInfo^.bmiHeader
                                 do begin
                                    if (biBitCount <= 8)
                                    then NumColors := (1 shl biBitCount)
                                    else NumColors := 0;
                                    LongWidth := (((biWidth * biBitCount) + 31) div 32) * 4;
                                    MaxBufSize := biHeight * LongWidth;
                                    MinBufSize := LongWidth;
                                 end;
                            end
                            else begin
                                 MinBufSize := 1024;
                                 MaxBufSize := 32768;
                            end;
                       end;
                       // else twRC := TWRC_FAILURE;

                       SetupMemXfer^.MinBufSize := MinBufSize;
                       SetupMemXfer^.MaxBufSize := MaxBufSize;
                       SetupMemXfer^.Preferred  := MaxBufSize;
                     end;
           else      begin
                       DSStatus.ConditionCode := TWCC_BADPROTOCOL;
                       twRC := TWRC_FAILURE;
                       DS_Error(E_SETUPMEMXFER, DSStatus.ConditionCode);
                     end;
           end;
  else     begin
             DSStatus.ConditionCode := TWCC_SEQERROR;
             twRC := TWRC_FAILURE;
             DS_Error(E_SETUPMEMXFER, DSStatus.ConditionCode);
           end;
  end;
  SetupMemXferMsg := twRC;
end; { End SetupMemXferMsg.                                                    }


{------------------------------------------------------------------------------}
{ FUNCTION: SetupFileXferMsg                                                   }
{                                                                              }
{ ARGS:    pTWMsg  message struct, same as list of parms in DS_ENTRY call      }
{                                                                              }
{ RETURNS: TWRC_SUCCESS                                                        }
{          TWRC_FAILURE                                                        }
{              TWCC_SEQERROR                                                   }
{              TWCC_BADVALUE                                                   }
{              TWCC_BADPROTOCOL        // not supportted                       }
{                                                                              }
{ NOTES:  Valid states: 4 -- 6.                                                }
{------------------------------------------------------------------------------}

function SetupFileXferMsg(pTWMsg : PTWMessage) : TW_UINT16;
var twRC    : TW_UINT16;
    pSetup  : pTW_SETUPFILEXFER;
    hFXFile : HFILE;
    FXof    : TOFSTRUCT;
begin
  LogMessage('SetupFileXferMsg');
  twRC := TWRC_SUCCESS;
  case DSState of
  STATE4,
  STATE5,
  STATE6 : case pTWMsg^.MSG of
           MSG_SET   : begin
                         pSetup := pTW_SETUPFILEXFER(pTWMsg^.pData);
                         lstrcpyn(FileName, pSetup^.FileName, SizeOf(FileName));

                         { Check that the file is valid.                       }
                         hFXFile := OpenFile(FileName, FXof, OF_EXIST);
                         if (integer(hFXFile) = HFILE_ERROR)
                         then begin
                              hFXFile := OpenFile(FileName, FXof, OF_CREATE);
                              if (integer(hFXFile) = HFILE_ERROR)
                              then begin
                                   DSStatus.ConditionCode := TWCC_BADVALUE;
                                   twRc := TWRC_FAILURE;
                                   DS_Error('Unable to create file <<DAT_SETUPFILEXFER>>',
                                            DSStatus.ConditionCode);
                              end
                              else _lclose(hFXFile);
                         end;
                       end;
           MSG_GETDEFAULT,
           MSG_GETCURRENT,
           MSG_GET   : begin
                         pSetup := pTW_SETUPFILEXFER(pTWMsg^.pData);
                         lstrcpyn(pSetup^.FileName,
                                  'c:\twain.bmp',
                                  SizeOf(pSetup^.FileName));
                         pSetup^.Format  := TWFF_BMP;
                         pSetup^.VRefNum := 0;
                       end;
           {
           Same as Else below.
           MSG_RESET : begin
                       end;
           }
           else        begin
                         DSStatus.ConditionCode := TWCC_BADPROTOCOL;
                         twRC := TWRC_FAILURE;
                         DS_Error(E_SETUPFILEXFER, DSStatus.ConditionCode);
                       end;
           end;
  else     begin
             DSStatus.ConditionCode := TWCC_SEQERROR;
             twRC := TWRC_FAILURE;
             DS_Error(E_SETUPFILEXFER, DSStatus.ConditionCode);
           end;
  end;
  SetupFileXferMsg := twRC;
end; { End SetupFileXferMsg.                                                   }


{------------------------------------------------------------------------------}
{ FUNCTION: XferGroupMsg                                                       }
{                                                                              }
{ ARGS:    pTWMsg  message struct, same as list of parms in DS_ENTRY call      }
{                                                                              }
{ RETURNS: status code                                                         }
{                                                                              }
{ NOTES:   Valid states: 4 -- 6.  Since only the Data Group is supported       }
{ under Version 1.0 of the specification the application should only use       }
{ MSG_GET.                                                                     }
{------------------------------------------------------------------------------}

function XferGroupMsg(pTWMsg : PTWMessage) : TW_UINT16;
var twRC : TW_UINT16;
begin
  LogMessage('XferGroupMsg');
  twRC := TWRC_SUCCESS;
  case DSState of
  STATE4,
  STATE5,
  STATE6 : case pTWMsg^.MSG of
           MSG_GET,
           MSG_GETDEFAULT,
           MSG_GETCURRENT,
           MSG_RESET      : pTW_UINT16(pTWMsg^.pData)^ := DG_IMAGE;
           MSG_SET        : ; { ONLY images are define in current spec so you  }
                              { can not change DG_ type.                       }
           else             begin
                              DSStatus.ConditionCode := TWCC_BADPROTOCOL;
                              twRC := TWRC_FAILURE;
                              DS_Error(E_XFERGROUP, DSStatus.ConditionCode);
                            end;
           end;
  else     begin
             DSStatus.ConditionCode := TWCC_SEQERROR;
             twRC := TWRC_FAILURE;
             DS_Error(E_XFERGROUP, DSStatus.ConditionCode);
           end;
  end;
  XferGroupMsg := twRC;
end; { End XferGroupMsg.                                                       }


{------------------------------------------------------------------------------}
{ FUNCTION: Palette8Msg                                                        }
{                                                                              }
{ ARGS:    pTWMsg  message struct, same as list of parms in DS_ENTRY call      }
{                                                                              }
{ RETURNS: status code                                                         }
{                                                                              }
{ NOTES:                                                                       }
{------------------------------------------------------------------------------}

function Palette8Msg(pTWMsg : PTWMessage) : TW_UINT16;
var twRc      : TW_UINT16;
    pPal      : pTW_PALETTE8;
    pDib      : PBITMAPINFO;
    index     : integer;
    NumColors : longint;
begin
  LogMessage('Palette8Msg');
  twRc := TWRC_SUCCESS;
  case DSState of
  STATE4,
  STATE5,
  STATE6 : begin
             case pTWMsg^.MSG of
             MSG_GET : begin
                         { MCM,                                                }
                         { Palette info is stored in pBmpInfo (ImageStream).   }
                         if (FormSource.ImageStream.Size = 0)
                         then FormSource.CopyBmpToStream;
                         if (FormSource.pBmpInfo <> Nil)
                         then begin
                              pDib      := FormSource.pBmpInfo;
                              NumColors := pDib^.bmiHeader.biClrUsed;
                              if (NumColors = 0)
                              then begin
                                   if (pDib^.bmiHeader.biBitCount <= 8)
                                   then NumColors := (1 shl pDib^.bmiHeader.biBitCount)
                                   else NumColors := 0;
                              end;
                              if (pDib^.bmiHeader.biClrUsed  =  0) and
                                 (pDib^.bmiHeader.biBitCount <= 8)
                              then pDib^.bmiHeader.biClrUsed := NumColors;

                              pPal := pTW_PALETTE8(pTWMsg^.pData);
                              pPal^.NumColors   := TW_UINT16(NumColors);
                              pPal^.PaletteType := TWPA_RGB;

                              for index := 0 to (NumColors - 1)
                              do begin
                                 with pDib^.bmiColors[index]
                                 do begin
                                    pPal^.Colors[index].Channel1 := rgbRed;
                                    pPal^.Colors[index].Channel2 := rgbGreen;
                                    pPal^.Colors[index].Channel3 := rgbBlue;
                                    pPal^.Colors[index].Index    := index;
                                 end;
                              end;
                         end;
                       end;
             MSG_GETDEFAULT,
             MSG_RESET,
             MSG_SET : begin
                       end;
             else      begin
                         DSStatus.ConditionCode := TWCC_BADPROTOCOL;
                         twRc := TWRC_FAILURE;
                         DS_Error(E_PALETTE8, DSStatus.ConditionCode);
                       end;
             end;
           end;
  else     begin
             DSStatus.ConditionCode := TWCC_SEQERROR;
             twRc := TWRC_FAILURE;
             DS_Error(E_PALETTE8, DSStatus.ConditionCode);
           end;
  end;
  Palette8Msg := twRc;
end; { End Palette8Msg.                                                        }


{------------------------------------------------------------------------------}
{ GrayResponseMsg -                                                            }
{------------------------------------------------------------------------------}

function GrayResponseMsg(pTWMsg : PTWMessage) : TW_UINT16;
var x       : TW_UINT16;
    TmpStr  : array[0..19] of char;
    RespNum : array[0..19] of char;
    Color   : array[0..19] of char;
begin
  LogMessage('GrayResponseMsg');
  for x := 0 to (16 - 1)
  do begin
     StrCopy(RespNum, 'Response[');
     Str(x, TmpStr);
     StrCat(RespNum, TmpStr);
     StrCat(RespNum, ']');

     StrCopy(Color, '{');
     Str(pTW_GRAYRESPONSE(pTWMsg^.pData)^.Response[x].Channel1, TmpStr);
     StrCat(Color, TmpStr);
     StrCat(Color, ',');
     Str(pTW_GRAYRESPONSE(pTWMsg^.pData)^.Response[x].Channel2, TmpStr);
     StrCat(Color, TmpStr);
     StrCat(Color, ',');
     Str(pTW_GRAYRESPONSE(pTWMsg^.pData)^.Response[x].Channel3, TmpStr);
     StrCat(Color, TmpStr);
     StrCat(Color, '}');

     WritePrivateProfileString('GrayResponse',
                               RespNum,
                               Color,
                               SourceControl);
  end;
  GrayResponseMsg := TWRC_SUCCESS;
end; { End GrayResponseMsg.                                                    }


{------------------------------------------------------------------------------}
{ RGBResponse - Note: Uses the GrayResponse structure which is identical to    }
{ the RGBRESPONSE structure in TWAIN 1.5.4                                     }
{------------------------------------------------------------------------------}

function RGBResponseMsg(pTWMsg : PTWMessage) : TW_UINT16;
var x  : TW_UINT16;
    TmpStr  : array[0..19] of char;
    RespNum : array[0..19] of char;
    Color   : array[0..19] of char;
begin
  LogMessage('RGBResponseMsg');
  for x := 0 to (16 - 1)
  do begin
     StrCopy(RespNum, 'Response[');
     Str(x, TmpStr);
     StrCat(RespNum, TmpStr);
     StrCat(RespNum, ']');

     StrCopy(Color, '{');
     Str(pTW_GRAYRESPONSE(pTWMsg^.pData)^.Response[x].Channel1, TmpStr);
     StrCat(Color, TmpStr);
     StrCat(Color, ',');
     Str(pTW_GRAYRESPONSE(pTWMsg^.pData)^.Response[x].Channel2, TmpStr);
     StrCat(Color, TmpStr);
     StrCat(Color, ',');
     Str(pTW_GRAYRESPONSE(pTWMsg^.pData)^.Response[x].Channel3, TmpStr);
     StrCat(Color, TmpStr);
     StrCat(Color, '}');

     WritePrivateProfileString('RGBResponse',
                               RespNum,
                               Color,
                               SourceControl);
  end;
  RGBResponseMsg := TWRC_SUCCESS;
end; { End RGBResponseMsg.                                                     }


{------------------------------------------------------------------------------}
{ CIEColorMsg -                                                                }
{------------------------------------------------------------------------------}

function CIEColorMsg(pTWMsg : PTWMessage) : TW_UINT16;
var AHandle : THandle;
    i       : integer;
begin
  LogMessage('CIEColorMsg');
  pTW_CIECOLOR(pTWMsg^.pData)^.ColorSpace      := TWPT_RGB;
  pTW_CIECOLOR(pTWMsg^.pData)^.LowEndian       := 0;
  pTW_CIECOLOR(pTWMsg^.pData)^.DeviceDependent := 1;
  pTW_CIECOLOR(pTWMsg^.pData)^.VersionNumber   := 0;

  { Transform stage ABC                                                        }
  pTW_CIECOLOR(pTWMsg^.pData)^.StageABC.Decode[0].StartIn := FloatToFIX32(0.1);

  pTW_CIECOLOR(pTWMsg^.pData)^.StageABC.Mix[0,0] := FloatToFIX32(0.0);
  pTW_CIECOLOR(pTWMsg^.pData)^.StageABC.Mix[0,1] := FloatToFIX32(0.1);
  pTW_CIECOLOR(pTWMsg^.pData)^.StageABC.Mix[0,2] := FloatToFIX32(0.2);
  pTW_CIECOLOR(pTWMsg^.pData)^.StageABC.Mix[1,0] := FloatToFIX32(1.0);
  pTW_CIECOLOR(pTWMsg^.pData)^.StageABC.Mix[1,1] := FloatToFIX32(1.1);
  pTW_CIECOLOR(pTWMsg^.pData)^.StageABC.Mix[1,2] := FloatToFIX32(1.2);
  pTW_CIECOLOR(pTWMsg^.pData)^.StageABC.Mix[2,0] := FloatToFIX32(2.0);
  pTW_CIECOLOR(pTWMsg^.pData)^.StageABC.Mix[2,1] := FloatToFIX32(2.1);
  pTW_CIECOLOR(pTWMsg^.pData)^.StageABC.Mix[2,2] := FloatToFIX32(2.2);

  { Transform Stage LMN                                                        }
  pTW_CIECOLOR(pTWMsg^.pData)^.StageLMN.Decode[0].StartIn := FloatToFIX32(1.1);

  pTW_CIECOLOR(pTWMsg^.pData)^.StageLMN.Mix[0,0] := FloatToFIX32(0.0);
  pTW_CIECOLOR(pTWMsg^.pData)^.StageLMN.Mix[0,1] := FloatToFIX32(0.1);
  pTW_CIECOLOR(pTWMsg^.pData)^.StageLMN.Mix[0,2] := FloatToFIX32(0.2);
  pTW_CIECOLOR(pTWMsg^.pData)^.StageLMN.Mix[1,0] := FloatToFIX32(1.0);
  pTW_CIECOLOR(pTWMsg^.pData)^.StageLMN.Mix[1,1] := FloatToFIX32(1.1);
  pTW_CIECOLOR(pTWMsg^.pData)^.StageLMN.Mix[1,2] := FloatToFIX32(1.2);
  pTW_CIECOLOR(pTWMsg^.pData)^.StageLMN.Mix[2,0] := FloatToFIX32(2.0);
  pTW_CIECOLOR(pTWMsg^.pData)^.StageLMN.Mix[2,1] := FloatToFIX32(2.1);
  pTW_CIECOLOR(pTWMsg^.pData)^.StageLMN.Mix[2,2] := FloatToFIX32(2.2);

  { CIEPoint                                                                   }
  pTW_CIECOLOR(pTWMsg^.pData)^.WhitePoint.X := FloatToFIX32(4.4);
  pTW_CIECOLOR(pTWMsg^.pData)^.WhitePoint.Y := FloatToFIX32(5.5);
  pTW_CIECOLOR(pTWMsg^.pData)^.WhitePoint.Z := FloatToFIX32(6.6);

  { CIEPoint                                                                   }
  pTW_CIECOLOR(pTWMsg^.pData)^.BlackPoint.X := FloatToFIX32(1.1);
  pTW_CIECOLOR(pTWMsg^.pData)^.BlackPoint.Y := FloatToFIX32(5.5);
  pTW_CIECOLOR(pTWMsg^.pData)^.BlackPoint.Z := FloatToFIX32(6.6);

  { CIEPoint                                                                   }
  pTW_CIECOLOR(pTWMsg^.pData)^.WhitePaper.X := FloatToFIX32(1.1);
  pTW_CIECOLOR(pTWMsg^.pData)^.WhitePaper.Y := FloatToFIX32(5.5);
  pTW_CIECOLOR(pTWMsg^.pData)^.WhitePaper.Z := FloatToFIX32(6.6);

  { CIEPoint                                                                   }
  pTW_CIECOLOR(pTWMsg^.pData)^.BlackInk.X := FloatToFIX32(1.1);
  pTW_CIECOLOR(pTWMsg^.pData)^.BlackInk.Y := FloatToFIX32(5.5);
  pTW_CIECOLOR(pTWMsg^.pData)^.BlackInk.Z := FloatToFIX32(6.6);

{ MCM ???                                                                      }
{ AHandle := GlobalAlloc(GHND, SizeOf(TW_FIX32) * 4);                          }
{ MCM ???                                                                      }
{ pTW_CIECOLOR(pTWMsg^.pData)^.Samples[0] := pTW_FIX32(GlobalLock(phandle));   }

  i := 0;
  pTW_CIECOLOR(pTWMsg^.pData)^.Samples[i] := FloatToFIX32(99.99);
  i := 1;
  pTW_CIECOLOR(pTWMsg^.pData)^.Samples[i] := FloatToFIX32(98.98);
  i := 2;
  pTW_CIECOLOR(pTWMsg^.pData)^.Samples[i] := FloatToFIX32(97.97);
  i := 3;
  pTW_CIECOLOR(pTWMsg^.pData)^.Samples[i] := FloatToFIX32(96.96);

{ MCM ???                                                                      }
{ GlobalUnlock(phandle);                                                       }

  CIEColorMsg := TWRC_SUCCESS;
end; { End CIEColorMsg.                                                        }


{------------------------------------------------------------------------------}
{ JPEGCompressionMsg -                                                         }
{------------------------------------------------------------------------------}

function JPEGCompressionMsg(pTWMsg : PTWMessage) : TW_UINT16;
begin
  LogMessage('JPEGCompressionMsg');
  JPEGCompressionMsg := TWRC_SUCCESS;
end; { End JPEGCompressionMsg.                                                 }


{------------------------------------------------------------------------------}
{ IsPaletteGray.                                                               }
{                                                                              }
{ MCM, This function is not a part of the original source code.                }
{                                                                              }
{------------------------------------------------------------------------------}

function IsPaletteGray : boolean;
var twRc      : TW_UINT16;
    pDib      : PBITMAPINFO;
    index     : integer;
    NumColors : longint;
    Reult     : boolean;
begin
  Result := False;
  if Assigned(FormSource)
  then begin
       if (FormSource.ImageStream.Size = 0)
       then FormSource.CopyBmpToStream;

       if (FormSource.pBmpInfo <> Nil)
       then begin
            pDib := FormSource.pBmpInfo;

            NumColors := pDib^.bmiHeader.biClrUsed;
            if (NumColors = 0)
            then if (pDib^.bmiHeader.biBitCount <= 8)
                 then NumColors := (1 shl pDib^.bmiHeader.biBitCount)
                 else NumColors := 0;

            index  := 0;
            Result := True;
            while (index < NumColors) and Result
            do begin
               with pDib^.bmiColors[index]
               do begin
                  if (rgbRed <> rgbGreen) or (rgbRed <> rgbBlue)
                  then Result := False
                  else inc(index);
               end;
            end;
            if (Index <> NumColors) and Result
            then Result := False;
       end;
  end;
  IsPaletteGray := Result;
end; { End IsPaletteGray.                                                      }


{------------------------------------------------------------------------------}
{ FUNCTION: ImageInfoMsg                                                       }
{                                                                              }
{ ARGS:    pTWMsg  message struct, same as list of parms in DS_ENTRY call      }
{                                                                              }
{ RETURNS: TWRC_SUCCESS                                                        }
{          TWRC_FAILURE                                                        }
{          TWCC_SEQERROR                                                       }
{ NOTES:  Valid states 6.  Fills in all fields of imageinfo struct.  Get       }
{ info on pending image transfer.                                              }
{ Fix the return codes, suppose to be pending, in which case it has already    }
{ been cleared for take off an none of the other message should occur at       }
{ this time !!!                                                                }
{------------------------------------------------------------------------------}

function ImageInfoMsg(pTWMsg : PTWMessage) : TW_UINT16;
var twRC         : TW_UINT16;
    i            : integer;
    pII          : pTW_IMAGEINFO;
    BitPerSample : integer;
    TmpStr       : string;
    GetSetStr    : string;
begin
  LogMessage('ImageInfoMsg');
  {$IFDEF DEBUG}
    case pTWMsg^.MSG of
    MSG_GET        : GetSetStr := ' \ MSG_GET';
    MSG_GETCURRENT : GetSetStr := ' \ MSG_GETCURRENT';
    MSG_GETDEFAULT : GetSetStr := ' \ MSG_GETDEFAULT';
    MSG_SET        : GetSetStr := ' \ MSG_SET';
    MSG_RESET      : GetSetStr := ' \ MSG_RESET';
    else GetSetStr := '- ERROR, BAD PROTECOL -';
    end;
    TmpStr := '' + GetSetStr;
    LogMessage(PChar(TmpStr));

    if (pTWMsg^.MSG = MSG_SET)
    then begin
         case pTW_CAPABILITY(pTWMsg^.pData)^.ConType of
         TWON_ARRAY       : GetSetStr := 'TWON_ARRAY';
         TWON_ENUMERATION : GetSetStr := 'TWON_ENUMERATION';
         TWON_ONEVALUE    : GetSetStr := 'TWON_ONEVALUE';
         TWON_RANGE       : GetSetStr := 'TWON_RANGE';
         else GetSetStr := '- ERROR -';
         end;
         GetSetStr := 'ConType: ' + GetSetStr;
         LogMessage(PChar(GetSetStr));
    end;
  {$ENDIF}

  twRC   := TWRC_SUCCESS;
  if (DSState = STATE6) 
  then begin
       case pTWMsg^.MSG of
       MSG_GET : begin
                   { MCM.                                                      }
                   { The original code is deleted. It was based on getting     }
                   {  info from a selected bitmap file.                        }

                   { MCM, 24.02.98                                             }
                   { Code in replacement of the original code.                 }
                   if (FormSource.ImageStream.Size = 0)
                   then FormSource.CopyBmpToStream;

                   if (FormSource.pBmpInfo <> Nil)
                   then begin
                        { set up local pointer to IMAGEINFO struct.            }
                        pII := pTW_IMAGEINFO(pTWMsg^.pData);

                        { use data in .BMP file to best fill the ImageInfo     }
                        { struct.                                              }
                        with FormSource.pBmpInfo^.bmiHeader
                        do begin
                           pII^.XResolution  := TW_FIX32(xRes.CurrentValue);
                           pII^.YResolution  := TW_FIX32(yRes.CurrentValue);
                           pII^.ImageWidth   := biWidth;
                           pII^.ImageLength  := biHeight;
                           pII^.BitsPerPixel := biBitCount * biPlanes;
                           pII^.Planar       := 0; { BMP image is chunky. }
                        end;
                        { MCM,                                                 }
                        { part of original code, but not used here.            }
                        { if (lpBifh^.biCompression = longint(0))              }
                        { then pII^.Compression  := TWCP_NONE;                 }

                        { MCM,                                                 }
                        { Images are never compressed.                         }
                        pII^.Compression  := Compression; { TWCP_NONE;         }

                        { clear first.                                         }
                        for i := 0 to 7
                        do pII^.BitsPerSample[i] := 0;

                        { assumes color .BMP are 1, 4, 8, 24 as spec. by MS.   }
                        if (pII^.BitsPerPixel <= 8)
                        then begin
                             pII^.SamplesPerPixel     := 1;
                             pII^.BitsPerSample[0]    := pII^.BitsPerPixel;
                             if IsPaletteGray
                             then pII^.PixelType      := TWPT_GRAY
                             else pII^.PixelType      := TWPT_PALETTE;
                        end
                        else begin
                             pII^.SamplesPerPixel     := 3;
                             with FormSource.pBmpInfo^.bmiHeader
                             do begin
                                case biBitCount of
                                1  : BitPerSample := 1;
                                4  : BitPerSample := 4;
                                8  : BitPerSample := 8;
                                16 : BitPerSample := 5;
                                24 : BitPerSample := 8;
                                32 : BitPerSample := 8;
                                end;
                             end;
                             for i := 0 to 2
                             do pII^.BitsPerSample[i] := BitPerSample;
                             pII^.PixelType           := TWPT_RGB;
                        end;
                   end;
                   twRC := TWRC_SUCCESS;
                 end;
       else      begin
                   DSStatus.ConditionCode := TWCC_BADPROTOCOL;
                   DS_Error(E_IMAGEINFOMSG, DSStatus.ConditionCode);
                   twRC := TWRC_FAILURE;
                 end;
       end;
  end
  else begin
       DSStatus.ConditionCode := TWCC_SEQERROR;
       twRC := TWRC_FAILURE;
       DS_Error(E_IMAGEINFOMSG, DSStatus.ConditionCode);
  end;
  ImageInfoMsg := twRC;
end; { End ImageInfoMsg.                                                       }


{------------------------------------------------------------------------------}
{ FUNCTION: ImageLayoutMsg                                                     }
{                                                                              }
{ ARGS:    pTWMsg  message struct, same as list of parms in DS_ENTRY call      }
{                                                                              }
{ RETURNS: status code                                                         }
{                                                                              }
{ NOTES:   This function is supported to the extent that it will               }
{ return success for each call.                                                }
{------------------------------------------------------------------------------}

function ImageLayoutMsg(pTWMsg : PTWMessage) : TW_UINT16;
var twRC       : TW_UINT16;
    pData      : pTW_IMAGELAYOUT;
    GetSetStr  : string;
    LogStr     : string;
begin
  twRC := TWRC_SUCCESS;
  LogMessage('ImageLayoutMsg');

  {$IFDEF DEBUG}
    case pTWMsg^.MSG of
    MSG_GET        : GetSetStr := ' \ MSG_GET';
    MSG_GETCURRENT : GetSetStr := ' \ MSG_GETCURRENT';
    MSG_GETDEFAULT : GetSetStr := ' \ MSG_GETDEFAULT';
    MSG_SET        : GetSetStr := ' \ MSG_SET';
    MSG_RESET      : GetSetStr := ' \ MSG_RESET';
    else GetSetStr := '- ERROR, BAD PROTECOL -';
    end;
    LogStr := '' + GetSetStr;
    LogMessage(PChar(LogStr));

    if (pTWMsg^.MSG = MSG_SET)
    then begin
         case pTW_CAPABILITY(pTWMsg^.pData)^.ConType of
         TWON_ARRAY       : GetSetStr := 'TWON_ARRAY';
         TWON_ENUMERATION : GetSetStr := 'TWON_ENUMERATION';
         TWON_ONEVALUE    : GetSetStr := 'TWON_ONEVALUE';
         TWON_RANGE       : GetSetStr := 'TWON_RANGE';
         else GetSetStr := '- ERROR -';
         end;
         GetSetStr := 'ConType: ' + GetSetStr;
         LogMessage(PChar(GetSetStr));
    end;
  {$ENDIF}
  pData := pTWMsg^.pData;
  case pTWMsg^.MSG of
  MSG_GET,
  MSG_GETDEFAULT,
  MSG_GETCURRENT : begin
                     if (FIX32ToFloat(ImageLayout.Frame.Left) = 0.0) and
                        (FIX32ToFloat(ImageLayout.Frame.Top) = 0.0) and
                        (FIX32ToFloat(ImageLayout.Frame.Right) = 0.0) and
                        (FIX32ToFloat(ImageLayout.Frame.Bottom) = 0.0)
                     then begin
                          if Assigned(FormSource)
                          then begin
                               if (FormSource.ImageStream.Size = 0)
                               then FormSource.CopyBmpToStream;
                               if Assigned(FormSource.pBmpInfo) and
                                  (FormSource.ImageStream.Size = 0)
                               then begin
                                    case UnitType[UnitEnum.CurrentIndex] of
                                    TWUN_INCHES,
                                    TWUN_CENTIMETERS,
                                    TWUN_PICAS       : begin
                                                         with ImageLayout,
                                                              FormSource.pBmpInfo^.bmiHeader
                                                         do begin
                                                            ImageLayout.Frame.Left   := FloatToFIX32(0.0);
                                                            ImageLayout.Frame.Top    := FloatToFIX32(0.0);
                                                            ImageLayout.Frame.Right  := FloatToFIX32(biWidth  / xRes.CurrentValue);
                                                            ImageLayout.Frame.Bottom := FloatToFIX32(biHeight / yRes.CurrentValue);
                                                         end;
                                                       end;
                                    TWUN_POINTS,
                                    TWUN_PIXELS      : begin
                                                         with ImageLayout,
                                                              FormSource.pBmpInfo^.bmiHeader
                                                         do begin
                                                            ImageLayout.Frame.Left.Long   := 0;
                                                            ImageLayout.Frame.Top.Long    := 0;
                                                            ImageLayout.Frame.Right.Long  := biWidth;
                                                            ImageLayout.Frame.Bottom.Long := biHeight;
                                                         end;
                                                       end;
                                    end;
                               end;
                          end;
                     end;
        	     pData^.Frame.Left     := ImageLayout.Frame.Left;
		     pData^.Frame.Top      := ImageLayout.Frame.Top;
		     pData^.Frame.Right    := ImageLayout.Frame.Right;
		     pData^.Frame.Bottom   := ImageLayout.Frame.Bottom;
		     pData^.DocumentNumber := ImageLayout.DocumentNumber;
		     pData^.PageNumber     := ImageLayout.PageNumber;
		     pData^.FrameNumber    := ImageLayout.FrameNumber;
                   end;
  MSG_SET        : begin
        	     ImageLayout.Frame.Left     := pData^.Frame.Left;
		     ImageLayout.Frame.Top      := pData^.Frame.Top;
		     ImageLayout.Frame.Right    := pData^.Frame.Right;
		     ImageLayout.Frame.Bottom   := pData^.Frame.Bottom;
		     ImageLayout.DocumentNumber := pData^.DocumentNumber;
		     ImageLayout.PageNumber     := pData^.PageNumber;
		     ImageLayout.FrameNumber    := pData^.FrameNumber;
                     case UnitType[UnitEnum.CurrentIndex] of
                     TWUN_INCHES,
                     TWUN_CENTIMETERS,
                     TWUN_PICAS       : begin
                                          with ImagePixFrame, pData^
                                          do begin
                                             Left   := Round(xRes.CurrentValue * FIX32ToFloat(Frame.Left));
                                             Top    := Round(yRes.CurrentValue * FIX32ToFloat(Frame.Top));
                                             Right  := Round(xRes.CurrentValue * FIX32ToFloat(Frame.Right));
                                             Bottom := Round(yRes.CurrentValue * FIX32ToFloat(Frame.Bottom));
                                          end;
                                        end;
                     TWUN_POINTS,
                     TWUN_PIXELS      : begin
                                          with ImagePixFrame
                                          do begin
                                             Left   := pData^.Frame.Left.Whole;
                                             Top    := pData^.Frame.Top.Whole;
                                             Right  := pData^.Frame.Right.Whole;
                                             Bottom := pData^.Frame.Bottom.Whole;
                                          end;
                                        end;
                     end;
                   end;
  MSG_RESET      : begin
                   end;
  else             begin
                     DSStatus.ConditionCode := TWCC_BADPROTOCOL;
                     DS_Error(E_IMAGELAYOUTMSG, DSStatus.ConditionCode);
                     twRC := TWRC_FAILURE;
                   end;
  end;
  ImageLayoutMsg := twRC;
end; { End ImageLayoutMsg


{------------------------------------------------------------------------------}
{ FUNCTION GetImageHandle.                                                     }
{ ARGS:    pTWMsg  message struct, same as list of parms in DS_ENTRY call      }
{ RETURNS: Handle to Image, otherwise 0.                                       }
{                                                                              }
{ MCM, This function is not a part of the original source code.                }
{                                                                              }
{                                                                              }
{------------------------------------------------------------------------------}

function GetImageHandle(A : PChar) : THandle;
type TLongType  = record
                  case Word of
        	  0  : (Ptr  : Pointer);
	          1  : (Long : Longint);
	 	  2  : (Lo   : Word;
	  	        Hi   : Word);
		  end;
     TAByte    = array[0..0] of byte;

var  hDIBs      : THandle;
     BitsS      : TLongType;
     BitsT      : TLongType;
     StartS     : TLongType;
     StartT     : TLongType;
     ToAddrS    : TLongType;
     ToAddrT    : TLongType;
     Count      : longint;
     BlockSize  : longint;
     LongWidth  : longint;
     NoColour   : word;
     HeaderSize : longint;
     i, j       : integer;
begin
  hDIBs  := 0;
  if (FormSource.ImageStream.Size = 0)
  then FormSource.CopyBmpToStream;

  if (FormSource.pBmpInfo <> Nil)
  then begin
       with FormSource.pBmpInfo^.bmiHeader
       do begin
          if (biBitCount <= 8)
          then NoColour := (1 shl biBitCount)
          else NoColour := 0;
          LongWidth := (((biWidth * biBitCount) + 31) div 32) * 4;
          if (biBitCount <= 8)
          then HeaderSize := SizeOf(TBitmapInfoHeader) +
                             (NoColour * SizeOf(TRGBQuad))
          else HeaderSize := SizeOf(TBitmapInfoHeader);
       end;

       { Create return bitmap.                                                 }
       hDIBs := GlobalAlloc(gmem_Moveable or gmem_Zeroinit,
                            FormSource.pBmpInfo^.bmiHeader.biSizeImage +
                            HeaderSize);
  end;

  if (hDIBs               <> 0)   and
     (FormSource.pBmp     <> Nil) and
     (FormSource.pBmpInfo <> Nil)
  then begin
       { MCM, Copy handle to bitmap.                                           }
       BitsS.Ptr := FormSource.pBmp;
       BitsT.Ptr := GlobalLock(hDIBs);

       { Place bitmap info in top of data block.                               }

       {$IFDEF WIN32}
         CopyMemory(BitsT.Ptr, FormSource.pBmpInfo, HeaderSize);
       {$ELSE}
         hmemcpy(BitsT.Ptr, FormSource.pBmpInfo, HeaderSize);
       {$ENDIF}

       if (BitsS.Ptr <> Nil) and
          (BitsT.Ptr <> Nil)
       then begin
            StartS.Long := 0;
            StartT.Long := HeaderSize;
            BlockSize   := FormSource.pBmpInfo.bmiHeader.biSizeImage;
            while (BlockSize > 0)
            do begin
               Count := BlockSize;
               if (Count > 1024)
               then Count := 1024;

               {$IFDEF WIN32}
                 { 32 Bit code                                                 }
                 { Calculate source address.                                   }
                 ToAddrS.Long := BitsS.Long + StartS.Long;

                 { Calculate target address.                                   }
                 ToAddrT.Long := BitsT.Long + StartT.Long;

                 { Move image data from source to target.                      }
                 CopyMemory(ToAddrT.Ptr, ToAddrS.Ptr, Count);
               {$ELSE}
                 { 16 Bit code                                                 }
                 { Calculate source address.                                   }
                 ToAddrS.Hi := BitsS.Hi + (StartS.Hi * Ofs(AHIncr));
                 ToAddrS.Lo := StartS.Lo;

                 { Calculate target address.                                   }
                 ToAddrT.Hi := BitsT.Hi + (StartT.Hi * Ofs(AHIncr));
                 ToAddrT.Lo := StartT.Lo;

                 { Move image data from source to target.                      }
                 hmemcpy(ToAddrT.Ptr, ToAddrS.Ptr, Count);
               {$ENDIF}

               { Increment source & target offset.                             }
               inc(StartS.Long, Count);
               inc(StartT.Long, Count);

               { Decrement remaining bytes to transfer.                        }
               dec(BlockSize, Count);
            end;
       end;
       GlobalUnlock(hDIBs);
  end
  else hDIBs := 0;
  GetImageHandle := hDIBs;
end; { End GetImageHandle.                                                     }


{------------------------------------------------------------------------------}
{ FUNCTION: ImageMemXferMsg                                                    }
{                                                                              }
{ ARGS:    pTWMsg  message struct, same as list of parms in DS_ENTRY call      }
{                                                                              }
{ RETURNS: TWRC_SUCCESS                                                        }
{          TWRC_FAILURE                                                        }
{              TWCC_BADVALUE                                                   }
{              TWCC_SEQERROR                                                   }
{          TWRC_XFERDONE                                                       }
{          TWRC_CANCEL                                                         }
{------------------------------------------------------------------------------}

function ImageMemXferMsg(pTWMsg : PTWMessage) : TW_UINT16;
var twRC        : TW_UINT16;
    pMem        : pTW_IMAGEMEMXFER;
    MaxRows     : TW_UINT32;
    RowsWritten : TW_UINT32;

    pDib        : PBITMAPINFO;
    DibStart    : TLongType;
    DibToAddr   : TLongType;
    DibBits     : TLongType;
    NumColors   : longint;
    LongWidth   : longint;
begin
  twRC := TWRC_SUCCESS;
  LogMessage('ImageMemXferMsg');

  if XferDone
  then begin
       { This is required so that the last time the transaction is processed,  }
       { the thunker won't try to write more data than is present.  Twacker    }
       { currently calls ImageMemXfer one time after all data has been         }
       { transfered.                                                           }
       pMem := pTW_IMAGEMEMXFER(pTWMsg^.pData);
       pMem^.BytesWritten := integer(0);

       { reset static variables.                                               }
       BytesOut  := 0;
       XferDone  := False;
       XferFirst := True;
       XferCount := 0;
       LogMessage('XferDone returned');
       twRC := TWRC_XFERDONE;
  end;

  if (twRC = TWRC_SUCCESS)
  then begin
       inc(XferCount);

       { abort if gets caught in an infinite loop.  The image should be        }
       { salvagable. 1000 should allow a very large number of transfers before }
       { quitting. If the image is broken on an uneven byte boundry, we may    }
       { get caught in a loop.                                                 }
       if (XferCount > 1000)
       then begin
            if (TimerOut = 999)
            then MessageBox(0, 'The source aborts a memory transfer',
                               'Sample Source', MB_OK)
            else TimedMsgBox(hDSInst, 0, 'The source aborts a memory transfer',
                                         'Sample Source',
                                         MB_OK, MB_OK, TimerOut);

            pMem := pTW_IMAGEMEMXFER(pTWMsg^.pData);
            pMem^.BytesWritten := longint(0);

            { reset static variables.                                          }
            BytesOut  := 0;
            XferDone  := False;
            XferFirst := True;
            XferCount := 0;
            LogMessage('XferDone Timed Out');
            twRC      := TWRC_XFERDONE;
       end;
  end;

  if (twRC = TWRC_SUCCESS)
  then begin
       case pTWMsg^.MSG of
       MSG_GET : begin
                   LogMessage('MSG_GET');

                   { Check that size is within the sources limits              }
                   pMem := pTW_IMAGEMEMXFER(pTWMsg^.pData);
                   if (FormSource.ImageStream.Size = 0)
                   then FormSource.CopyBmpToStream;
                   if (MinBufSize > pMem^.Memory.Length) {or
                      (MaxBufSize < pMem^.Memory.Length) }
                   then begin
                        twRC := TWRC_FAILURE;
                        exit;
                   end;

                   pDib  := FormSource.pBmpInfo;
                   if (pDib <> Nil)
                   then begin
                        if (BytesOut = 0)
                        then FlipBitmap(FormSource.pBmp,  FormSource.pBmpInfo);
                        with FormSource.pBmpInfo.bmiHeader
                        do begin
                           if (biBitCount <= 8)
                           then NumColors := (1 shl biBitCount)
                           else NumColors := 0;
                           LongWidth := (((biWidth * biBitCount) + 31) div 32) * 4;
                        end;
                        { increment the pointer to the bitmap to the correct   }
                        { offset within the data area DibPtr will then point   }
                        { to the start of the current block of data to be      }
                        { copied.                                              }
                        DibBits.Ptr   := pDib;

                        DibStart.Long := SizeOf(TBITMAPINFOHEADER);
                        DibStart.Long := DibStart.Long +
                                         NumColors * SizeOf(TRGBQUAD);
                        DibStart.Long := DibStart.Long + BytesOut;

                        {$IFDEF WIN32}
                          { 32 Bit code.                                       }
                          DibToAddr.Long := DibBits.Long + DibStart.Long;
                        {$ELSE}
                          { 16 Bit code.                                       }
                          DibToAddr.Hi  := DibBits.Hi + (DibStart.Hi * Ofs(AHIncr));
                          DibToAddr.Lo  := DibStart.Lo;
                        {$ENDIF}

                        LogMessage('Incremented pointer');

                        { Set the attributes of the xfer.                      }
                        pMem^.Compression := TWCP_NONE;

                        pMem^.BytesPerRow := LongWidth;

                        { Need to check that at least one entire strip can be  }
                        { transferred within the bounds of the buffer size.    }
                        { If the buffer is smaller than one strip, a tile      }
                        { transfer will have to be executed. This code will    }
                        { fail since tile is not supported.                    }
                        if (pMem^.BytesPerRow > pMem^.Memory.Length)
                        then begin
                             { need to set condition code also.                }
                             MessageBox(0, 'Buffer is too small for strip transfers',
                                           '', MB_OK);
                             ImageMemXferMsg := TWRC_FAILURE;
                             exit;
                        end;

                        pMem^.Columns := pDib^.bmiHeader.biWidth;
                        pMem^.Rows    := pDib^.bmiHeader.biHeight;

                        pMem^.XOffset := 0;
                        if (pMem^.BytesPerRow > 0)
                        then pMem^.YOffset := BytesOut div pMem^.BytesPerRow + 1
                        else pMem^.YOffset := 1;

                        LogMessage('Set structure Y:');
                        LogMessage(PChar(IntToStr(pMem^.YOffset)));

                        { Determine the max number of rows that can be passed  }
                        { this time thru the transfer must end on a strip (row)}
                        { boundry.                                             }
                        MaxRows := pMem^.Memory.Length div pMem^.BytesPerRow;

                        { Determine the number of rows of the bitmap that have }
                        { been written thus far                                }
                        RowsWritten := BytesOut div pMem^.BytesPerRow;

                        { If the bitmap or the remaining bitmap is smaller     }
                        { than the memory allocation.                          }
                        if (MaxRows >= (pMem^.Rows - RowsWritten))
                        then pMem^.BytesWritten := pMem^.BytesPerRow *
                                                  (pMem^.Rows - RowsWritten)
                        else begin
                             { It will use the entire buffer area.             }
                             pMem^.BytesWritten := pMem^.BytesPerRow * MaxRows;
                        end;
                        { pMem^.Rows = number of rows to transfer per call.    }
                        pMem^.Rows := pMem^.BytesWritten div pMem^.BytesPerRow;

                        BytesOut := BytesOut + pMem^.BytesWritten;

                        { Determine the number of rows of the bitmap that have }
                        { been written thus far                                }
                        RowsWritten := BytesOut div pMem^.BytesPerRow;

                        { if the bytes written total are in excess of the      }
                        { bitmap image size, we are done.                      }
                        { if (BytesOut >= pDib^.bmiHeader.biSizeImage)         }
                        {     done = TRUE;                                     }
                        if ((pDib^.bmiHeader.biHeight{pMem^.Rows} - RowsWritten) = 0)
                        then begin
                             XferDone := True;
                             twRC     := TWRC_XFERDONE;
                             LogMessage('Memory buffer Success');
                             {$IFDEF TCITWAIN}
{ Deletes image premature.
                               if Assigned(FormSource)
                               then FormSource.ImageChanged(FormSource);
}                               
                             {$ENDIF}
                        end
                        else twRC := TWRC_SUCCESS;

                        { Copy memory to the shared pointer                    }
                        { (owned by the application)                           }
                        {$IFDEF WIN32}
                          { 32 Bit code.                                       }

                          CopyMemory(pMem^.Memory.TheMem,
                                     DibToAddr.Ptr,
                                     pMem^.BytesWritten);
                        {$ELSE}
                          { 16 Bit code.                                       }
                          hmemcpy(pMem^.Memory.TheMem,
                                  DibToAddr.Ptr,
                                  pMem^.BytesWritten);
                        {$ENDIF}
                        LogMessage('Copied buffer memory');
                   end
                   else begin { MCM, Added in case returned hIBmp = 0.         }
                        pMem^.BytesWritten     := longint(0);
                        DSStatus.ConditionCode := TWCC_BADVALUE;
                        twRC := TWRC_FAILURE;
                        DS_Error(E_IMAGELAYOUTMSG, DSStatus.ConditionCode);
                   end;
                 end;
       else      begin
                   DSStatus.ConditionCode := TWCC_BADPROTOCOL;
                   twRc := TWRC_FAILURE;
                   DS_Error(E_IMAGELAYOUTMSG, DSStatus.ConditionCode);
                 end;
       end;
  end;
  ImageMemXferMsg := twRC;
end; { End ImageMemXferMsg.                                                    }


{------------------------------------------------------------------------------}
{ FUNCTION: ImageFileXferMsg                                                   }
{                                                                              }
{ ARGS:    pTWMsg  message struct, same as list of parms in DS_ENTRY call      }
{                                                                              }
{ RETURNS: TWRC_DONE                                                           }
{          TWRC_FAILURE                                                        }
{              TWCC_BADVALUE                                                   }
{              TWCC_SEQERROR                                                   }
{          TWRC_XFERDONE                                                       }
{              TWCC_CANCEL                                                     }
{                                                                              }
{ NOTES:   Valid states: 6 -- 7.  I do not support file transfers.             }
{------------------------------------------------------------------------------}

function ImageFileXferMsg(pTWMsg : PTWMessage) : TW_UINT16;
var twRC : TW_UINT16;
begin
  LogMessage('ImageFileXferMsg');
  twRC := TWRC_XFERDONE;
  case pTWMsg^.MSG of
  MSG_GET : begin
              // it has already been determined that the file exists and
              // can be opened, it is safe to copy it
              FormSource.Bitmap.SaveToFile(FileName);
              twRc := TWRC_XFERDONE;
            end;
  else      begin
              DSStatus.ConditionCode := TWCC_BADPROTOCOL;
              DS_Error(E_IMAGELAYOUTMSG, DSStatus.ConditionCode);
              twRC := TWRC_FAILURE;
            end;
  end;
  ImageFileXferMsg := twRC;
end; { End ImageFileXferMsg.                                                   }


{------------------------------------------------------------------------------}
{ FUNCTION: ImageNativeXferMsg                                                 }
{                                                                              }
{ ARGS:    pTWMsg  message struct, same as list of parms in DS_ENTRY call      }
{                                                                              }
{ RETURNS: TWRC_SUCCESS                                                        }
{          TWRC_FAILURE                                                        }
{              TWCC_BADVALUE                                                   }
{              TWCC_SEQERROR                                                   }
{          TWRC_XFERDONE                                                       }
{              TWCC_CANCEL                                                     }
{              TWCC_BADPROTOCOL         // debug only                          }
{                                                                              }
{ NOTES:  Valid states: 6 -- 7.  I only support native transfers.              }
{------------------------------------------------------------------------------}

function ImageNativeXferMsg(pTWMsg : PTWMessage) : TW_UINT16;
type LongType  = record
                 case Word of
		 0  : (Ptr  : Pointer);
		 1  : (Long : Longint);
		 2  : (Lo   : Word;
		       Hi   : Word);
		 end;
     TAByte    = array[0..0] of byte;
var  twRC       : TW_UINT16;
     Temp       : TW_UINT32;
     Bits       : LongType;
     Count      : Longint;
     New_Count  : longint;
     Old_Count  : longint;
     Start      : LongType;
     StartTmp   : LongType; { Temp array.                                   }
     ToAddr     : LongType;
     ToAddrTmp  : LongType; { Temp array.                                   }
     CDown      : LongType;

     hBits      : THandle;
     pBits      : pointer;
     hInfo      : THandle;
     pInfo      : pBitmapInfo;

     bitCount   : word;
     NoColour   : word;
     LongWidth  : longint;
     HeaderSize : longint;
     i, j       : integer;
begin
  LogMessage('ImageNativeXferMsg');
  twRC := TWRC_SUCCESS;
  if ((DSState = STATE6) or (DSState = STATE7))
  then begin
       case pTWMsg^.MSG of
       MSG_GET : begin // return hDib in low word.
                   pTW_UINT32(pTWMsg^.pData)^ := GetImageHandle(TWAINFileName);
                   twRc := TWRC_XFERDONE;
                 end;
       else      begin
                   DSStatus.ConditionCode := TWCC_BADPROTOCOL;
                   twRC := TWRC_FAILURE;
                   DS_Error(E_IMAGENATIVEXFERMSG, DSStatus.ConditionCode);
                 end;
       end;
  end
  else begin
       DSStatus.ConditionCode := TWCC_SEQERROR;
       DS_Error(E_IMAGENATIVEXFERMSG, DSStatus.ConditionCode);
       twRC := TWRC_FAILURE;
  end;
  ImageNativeXferMsg := twRC;
end; { End ImageNativeXferMsg.                                                 }


{------------------------------------------------------------------------------}
{ FUNCTION: NotifyCloseDSReq                                                   }
{                                                                              }
{ ARGS:    none                                                                }
{                                                                              }
{ RETURNS: Indirectly returns a MSG_CLOSEDSREQ to the Application, thru        }
{          DSM_Entry call.                                                     }
{------------------------------------------------------------------------------}

procedure NotifyCloseDSReq;
var hDSMDLL     : THandle;
    lpDSM_Entry : TDSM_Entry_Proc; { Entry point to the SM.                    }
begin
  LogMessage('NotifyCloseDSReq');
{$IFDEF WIN32}
   // SDH - 02/03/95 - New Twain32 dll name for thunker.
   hDSMDLL := LoadLibrary('TWAIN_32.DLL');
   if (hDSMDLL <> 0)
{$ELSE} // GR: 16 bit version
   hDSMDLL := LoadLibrary('TWAIN.DLL');
   if (hDSMDLL <> 0)
{$ENDIF}
   then begin
        @lpDSM_Entry := GetProcAddress(hDSMDLL, 'DSM_Entry');
        if (@lpDSM_Entry <> Nil)
        then begin
             lpDSM_Entry(@dsIdentity,
                         @appIdentity,
                         TW_UINT32(DG_CONTROL),
                         DAT_NULL,
                         MSG_CLOSEDSREQ,
                         TW_MEMREF(Nil));
        end;
        FreeLibrary(hDSMDLL);
   end;
end; { End NotifyCloseReq.                                                     }


{------------------------------------------------------------------------------}
{ FUNCTION: NotifyXferReady                                                    }
{                                                                              }
{ ARGS:    none                                                                }
{                                                                              }
{ RETURNS: Indirectly returns a MSG_XFERREADY to the Application, thru         }
{          DSM_Entry call.                                                     }
{------------------------------------------------------------------------------}

procedure NotifyXferReady;
var hDSMDLL     : THandle;
    lpDSM_Entry : TDSM_Entry_Proc; // Entry point to the SM.
begin
  LogMessage('NotifyXferReady');
  // go to xferready state.
  DSState := STATE6;

{$IFDEF WIN32}
   // SDH - 02/03/95 - New Twain32 dll name for thunker.
   hDSMDLL := LoadLibrary('TWAIN_32.DLL');
   if (hDSMDLL <> 0)
{$ELSE} // GR: 16 bit version
   hDSMDLL := LoadLibrary('TWAIN.DLL');
   if (hDSMDLL <> 0)
{$ENDIF}
   then begin
        @lpDSM_Entry := GetProcAddress(hDSMDLL, 'DSM_Entry');
        if (@lpDSM_Entry <> Nil)
        then begin
             lpDSM_Entry(@dsIdentity,
                         @appIdentity,
                         TW_UINT32(DG_CONTROL),
                         DAT_NULL,
                         MSG_XFERREADY,
                         TW_MEMREF(Nil));
        end;
        FreeLibrary(hDSMDLL);
   end;
end; { End NotifyXferReady.                                                    }


{------------------------------------------------------------------------------}
{ FlipBitMap - Takes a memory transfer buffer and changes it to a DIB format   }
{                                                                              }
{    i.e.    Memory Format                                                     }
{                         1  2  3  4  5                                        }
{                         6  7  8  9 10                                        }
{                        11 12 13 14 15                                        }
{                                                                              }
{            DIB bitmap Format                                                 }
{                        11 12 13 14 15                                        }
{                         6  7  8  9 10                                        }
{                         1  2  3  4  5                                        }
{                                                                              }
{    Memory RGBQuad order: RGB                                                 }
{    Windows DIB RGBQuad order: BGR                                            }
{                                                                              }
{ MCM 210996, This procedure was re-written in order to use less memory while  }
{ flipping the bitmap.                                                         }
{                                                                              }
{------------------------------------------------------------------------------}

procedure FlipBitmap(pBmp     : Pointer;
                     pBmpInfo : PBitmapInfo);
var hBmpT       : THandle;
    pBmpT       : PVectorB;
    hBmpB       : THandle;
    pBmpB       : PVectorB;

    Height      : longint;
    SizeImage   : longint;
    BitCount    : word;
    NumColors   : integer;
    ClrUsed     : longint;
    Offset      : longint;
    pixels      : TW_UINT16;
    items       : TW_INT32;
    i, j        : TW_UINT32;
    SaveRed     : byte;

    { MCM 210996, Variables needed to handle large bitmaps.                    }
    LongWidth   : longint;
    StartT      : TLongType;
    ToAddrT     : TLongType;
    BitsT       : TLongType;

    StartB      : TLongType;
    ToAddrB     : TLongType;
    BitsB       : TLongType;
begin
  Height     := pBmpInfo^.bmiHeader.biHeight;
  SizeImage  := pBmpInfo^.bmiHeader.biSizeImage;
  BitCount   := pBmpInfo^.bmiHeader.biBitCount;
  ClrUsed    := pBmpInfo^.bmiHeader.biClrUsed;
  LongWidth  := (((pBmpInfo^.bmiHeader.biWidth * bitCount) + 31) div 32) * 4;

  hBmpT := GlobalAlloc(GHND, LongWidth);
  hBmpB := GlobalAlloc(GHND, LongWidth);
  if (hBmpT <> 0) and
     (hBmpB <> 0)
  then begin
       pBmpT     := GlobalLock(hBmpT);
       pBmpB     := GlobalLock(hBmpB);

       BitsT.Ptr := pBmpInfo;
       BitsB.Ptr := pBmpInfo;

       if (pBmpInfo^.bmiHeader.biBitCount <= 8)
       then NumColors := (1 shl pBmpInfo^.bmiHeader.biBitCount)
       else NumColors := 0;

       { Calculate Offset to start of the bitmap data.                         }
       Offset    := SizeOf(TBitmapInfoHeader);
       Offset    := Offset + NumColors * SizeOf(TRGBQUAD);

       StartT.Long := Offset;
       StartB.Long := Offset + (LongWidth * (Height - 1));

       { For each line.                                                        }
       for i := 0 to ((Height - 1) div 2)
       do begin
          {$IFDEF WIN32}
            { 32 Bit code.                                                     }
            ToAddrT.Long := BitsT.Long + StartT.Long;
            ToAddrB.Long := BitsB.Long + StartB.Long;

            CopyMemory(pBmpT, ToAddrT.Ptr, LongWidth);
            CopyMemory(pBmpB, ToAddrB.Ptr, LongWidth);

            CopyMemory(ToAddrB.Ptr, pBmpT, LongWidth);
            CopyMemory(ToAddrT.Ptr, pBmpB, LongWidth);

            StartT.Long := StartT.Long + LongWidth;
            StartB.Long := StartB.Long - LongWidth;

          {$ELSE}
            { 16 Bit code.                                                     }
            ToAddrT.Hi := BitsT.Hi + (StartT.Hi * Ofs(AHIncr));
            ToAddrT.Lo := StartT.Lo;

            ToAddrB.Hi := BitsB.Hi + (StartB.Hi * Ofs(AHIncr));
            ToAddrB.Lo := StartB.Lo;
            hmemcpy(pBmpT, ToAddrT.Ptr, LongWidth);
            hmemcpy(pBmpB, ToAddrB.Ptr, LongWidth);

            hmemcpy(ToAddrB.Ptr, pBmpT, LongWidth);
            hmemcpy(ToAddrT.Ptr, pBmpB, LongWidth);

            StartT.Long := StartT.Long + LongWidth;
            StartB.Long := StartB.Long - LongWidth;
          {$ENDIF}
       end;

       { Flip RGB color table.                                                 }
       if (BitCount = 24)
       then begin
            Pixels      := TW_UINT16(pBmpInfo^.bmiHeader.biWidth);
            StartT.Long := Offset;
            for items := 0 to (Height - 1)
            do begin
               {$IFDEF WIN32}
                 { 32 Bit code.                                                }
                 ToAddrT.Long := BitsT.Long + StartT.Long;
                 CopyMemory(pBmpT, ToAddrT.Ptr, LongWidth);
                 for i := 0 to (Pixels - 1)
                 do begin
                    { Switch Red byte and Blue byte.                           }
                    j := i * 3;
                    SaveRed     := pBmpT^[j];
                    pBmpT^[j]   := pBmpT^[j+2];
                    pBmpT^[j+2] := SaveRed;
                 end;
                 CopyMemory(ToAddrT.Ptr, pBmpT, LongWidth);
                 StartT.Long := StartT.Long + LongWidth;

               {$ELSE}
                 { 16 Bit code.                                                }
                 ToAddrT.Hi := BitsT.Hi + (StartT.Hi * Ofs(AHIncr));
                 ToAddrT.Lo := StartT.Lo;
                 hmemcpy(pBmpT, ToAddrT.Ptr, LongWidth);
                 for i := 0 to (Pixels - 1)
                 do begin
                    { Switch Red byte and Blue byte.                           }
                    j := i * 3;
                    SaveRed     := pBmpT^[j];
                    pBmpT^[j]   := pBmpT^[j+2];
                    pBmpT^[j+2] := SaveRed;
                 end;
                 hmemcpy(ToAddrT.Ptr, pBmpT, LongWidth);
                 StartT.Long := StartT.Long + LongWidth;
               {$ENDIF}
          end;
       end;

       { Unlock memory.                                                        }
       GlobalUnlock(hBmpT);
       GlobalUnlock(hBmpB);
  end;

  { Free.                                                                      }
  if (hBmpT <> 0)
  then GlobalFree(hBmpT);
  if (hBmpB <> 0)
  then GlobalFree(hBmpB);
end; { End FlipBitmap.                                                         }


{------------------------------------------------------------------------------}
{ FUNCTION: DS_Error                                                           }
{                                                                              }
{ ARGS:    *state  ptr to char string of error message                         }
{          code    condition code TWCC_xxxx                                    }
{                                                                              }
{ RETURNS: none, does output a message box                                     }
{------------------------------------------------------------------------------}

procedure DS_Error(state : PChar; code : integer);
var details : array[0..179] of char;
begin
  { Put in detail of current operation.                                        }
  lstrcpy(details, state);
  case code of
  TWCC_BADCAP         : lstrcat(details, 'Bad capabilities error.');
  TWCC_BADPROTOCOL    : lstrcat(details, 'Bad protocol error.');
  TWCC_BADVALUE       : lstrcat(details, 'Bad value error.');
  TWCC_BUMMER         : lstrcat(details, 'Bummer error.');
  TWCC_OPERATIONERROR : lstrcat(details, 'An operation error just occurred.');
  TWCC_LOWMEMORY      : lstrcat(details, 'Low Memory Condition encountered, free up some memory and try again.');
  TWCC_MAXCONNECTIONS : lstrcat(details, 'Source is connected to max possible apps.');
  TWCC_NODS           : lstrcat(details, 'No Data Source found error.');
  TWCC_SEQERROR       : lstrcat(details, 'DG DAT MSG is out of expected sequence.');
  TWCC_SUCCESS        : lstrcat(details, 'Success message from detailed error request ??');
                   else lstrcat(details, 'Unknown Error value in condition code.');
  end;
  { Timerout is set to 999 initially and only reset with the autotests.        }
  if (TimerOut = 999)
  then MessageBox(0,  details, 'Source Error', MB_OK)
  else TimedMsgBox(hDSInst, hAppWnd, details, 'Sample Source Error',
                            MB_OK, MB_OK, TimerOut);
  { Rules say to return a CLOSEDSREQ for fatal errors.                         }
end; { End DS_Error.                                                           }


{------------------------------------------------------------------------------}
{ CopyFileEx - copies file1 to file2.  This function is provided for backwards }
{ compatibility to 16 bit Windows.  The CopyFile function is not supported at  }
{ 16 bits                                                                      }
{------------------------------------------------------------------------------}

{$IFNDEF WIN32}
  { 16 Bit code.                                                               }

function CopyFileEx(Source : PChar;
                    Dest   : PChar;
                    Notify : bool) : bool;
var hSource  : HFile;
    hDest    : HFile;
    Count    : word;
    out      : word;
    OFSTRUCT : TOFSTRUCT;
    buffer   : array[0..2040] of char;
begin
  hSource := OpenFile(Source, OFSTRUCT, OF_READ);
  if (integer(hSource) = HFILE_ERROR)
  then begin
       CopyFileEx := False;
       exit;
  end;
  hDest := OpenFile(Dest, OFSTRUCT, OF_CREATE);
  if (integer(hDest) = HFILE_ERROR)
  then begin
       CopyFileEx := False;
       exit;
  end;

  Count := 2040;
  Out := _lread(hSource, Buffer, Count);
  while(Out > 0)
  do begin
     _lwrite(hDest, Buffer, Out);
     Out := _lread(hSource, Buffer, Count);
  end;

  { Close files and release memory.                                            }
  _lclose(hSource);
  _lclose(hDest);

  CopyFileEx := True;
end; { End CopyFileEx.                                                         }

{$ENDIF}

{------------------------------------------------------------------------------}
{ SetSourceControl -- the source control filename is placed in a temp file     }
{ at the time when the SET SourceControl macro call is made by the app or a    }
{ the send dialog is initiated.  The temp file is then erased here after       }
{ being read.                                                                  }
{------------------------------------------------------------------------------}

function SetSourceControl : bool;
var TempDir  : array[0..MAX_TWPATH] of char;
    hAFile   : HFILE;
    OFSTRUCT : TOFSTRUCT;
begin
  SetSourceControl := True;

  { Set the control file path and filename -- this must exist.                 }
  GetWindowsDirectory(TempDir, SizeOf(TempDir));
  lstrcat(TempDir,'\source.$$$');

  {$IFDEF WIN32}
    { 32 Bit code.                                                             }
    hAfile := OpenFile(TempDir, OFSTRUCT, OF_READ);
    if (integer(hAfile) <> HFILE_ERROR)
    then begin
         FillChar(SourceControl, MAX_TWPATH, #0);
         _lread(hAfile, @SourceControl, MAX_TWPATH);
         _lclose(hAfile);
    end
  {$ELSE}
    { 16 Bit code.                                                             }
    hAfile := OpenFile(TempDir, OFSTRUCT, OF_READ);
    if (integer(hAfile) <> HFILE_ERROR)
    then begin
         FillChar(SourceControl, MAX_TWPATH, #0);
         _lread(hAfile, SourceControl, MAX_TWPATH);
         _lclose(hAfile);
    end
  {$ENDIF}
  else begin
       if (SourceControl[0] = #0) { First time in.                             }
       then begin
            GetWindowsDirectory(TempDir, SizeOf(TempDir));
            lstrcat(TempDir, '\default.src');
            if (integer(OpenFile(TempDir, OFSTRUCT, OF_EXIST)) <> HFILE_ERROR)
            then begin
                 lstrcpyn(SourceControl, TempDir, SizeOf(TempDir));
            end
            else begin
                 MessageBox(0, 'Unable to find default Source Control', '',
                               MB_OK);
                 SetSourceControl := False;
                 exit;
            end;
       end;
  end;
end; { End SetSourceControl.                                                   }


{------------------------------------------------------------------------------}
{ LogMessage -                                                                 }
{------------------------------------------------------------------------------}

procedure LogMessage(Msg : PChar);
var hAFile   : HFILE;
    OFSTRUCT : TOFSTRUCT;
begin
  {$IFDEF DEBUG}
    if (integer(OpenFile('c:\twsrc.log', OFSTRUCT, OF_EXIST)) = HFILE_ERROR)
    then hAFile := OpenFile('c:\twsrc.log', OFSTRUCT, OF_CREATE or OF_WRITE)
    else hAFile := OpenFile('c:\twsrc.log', OFSTRUCT, OF_WRITE);

    _llseek(hAFile, 0, FILE_END);
    _lwrite(hAFile, Msg, lstrlen(Msg));
    _lwrite(hAfile, CR_LF, lstrlen(CR_LF));
    _lclose(hAFile);
  {$ENDIF}
end; { End LogMessage.                                                         }


{------------------------------------------------------------------------------}
{ ShowTW_ID -                                                                  }
{------------------------------------------------------------------------------}

procedure ShowTW_ID(hAppWnd : hWnd; twID : TW_IDENTITY; Text : PChar);
var Details  : array[0..255] of char;
    Details2 : array[0..1024] of char;
begin
  {$IFDEF WIN32}
    { 32 Bit code.                                                             }
    StrFmt(Details2, 'Id: %d', [twID.Id]);
    StrCat(Details2, CR_LF);
    StrFmt(Details, 'TW_VERSION.MajorNum: %d', [twID.Version.MajorNum]);
    StrCat(Details2, Details);
    StrCat(Details2, CR_LF);
    StrFmt(Details, 'TW_VERSION.MinorNum: %d', [twID.Version.MinorNum]);
    StrCat(Details2, Details);
    StrCat(Details2, CR_LF);
    StrFmt(Details, 'TW_VERSION.Language: %d', [twID.Version.Language]);
    StrCat(Details2, Details);
    StrCat(Details2, CR_LF);
    StrFmt(Details, 'TW_VERSION.Country: %d', [twID.Version.Country]);
    StrCat(Details2, Details);
    StrCat(Details2, CR_LF);
    StrCat(Details2, 'TW_VERSION.Info: ');
    StrCat(Details2, twID.Version.Info);
    StrCat(Details2, CR_LF);
    StrFmt(Details, 'ProtocolMajor: %d', [twID.ProtocolMajor]);
    StrCat(Details2, Details);
    StrCat(Details2, CR_LF);
    StrFmt(Details, 'ProtocolMinor: %d', [twID.ProtocolMinor]);
    StrCat(Details2, Details);
    StrCat(Details2, CR_LF);
    StrFmt(Details, 'SupportedGroups: %d', [twID.SupportedGroups]);
  {$ELSE}
    { 16 Bit code.                                                             }
    wvsprintf(Details2, 'Id: %d', twID.Id);
    StrCat(Details2, CR_LF);
    wvsprintf(Details, 'TW_VERSION.MajorNum: %d', twID.Version.MajorNum);
    StrCat(Details2, Details);
    StrCat(Details2, CR_LF);
    wvsprintf(Details, 'TW_VERSION.MinorNum: %d', twID.Version.MinorNum);
    StrCat(Details2, Details);
    StrCat(Details2, CR_LF);
    wvsprintf(Details, 'TW_VERSION.Language: %d', twID.Version.Language);
    StrCat(Details2, Details);
    StrCat(Details2, CR_LF);
    wvsprintf(Details, 'TW_VERSION.Country: %d',  twID.Version.Country);
    StrCat(Details2, Details);
    StrCat(Details2, CR_LF);
    StrCat(Details2, 'TW_VERSION.Info: ');
    StrCat(Details2, twID.Version.Info);
    StrCat(Details2, CR_LF);
    wvsprintf(Details, 'ProtocolMajor: %d',       twID.ProtocolMajor);
    StrCat(Details2, Details);
    StrCat(Details2, CR_LF);
    wvsprintf(Details, 'ProtocolMinor: %d',       twID.ProtocolMinor);
    StrCat(Details2, Details);
    StrCat(Details2, CR_LF);
    wvsprintf(Details, 'SupportedGroups: %d',     twID.SupportedGroups);
  {$ENDIF}

  StrCat(Details2, Details);
  StrCat(Details2, CR_LF);
  StrCat(Details2, 'Manufacturer: ');
  StrCat(Details2, twID.Manufacturer);
  StrCat(Details2, CR_LF);
  StrCat(Details2, 'ProductFamily: ');
  StrCat(Details2, twID.ProductFamily);
  StrCat(Details2, CR_LF);
  StrCat(Details2, 'ProductName: ');
  StrCat(Details2, twID.ProductName);
  StrCat(Details2, CR_LF);

{  MessageBox(hImageDlg, Details2, Text, MB_OK);                               }
end; { End ShowTW_ID.                                                          }


end. { End Unit twd_prot.                                                      }

